package edu.georgetown.gucs.dictionary;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.net.UnknownHostException;
import java.text.DecimalFormat;
import java.util.Comparator;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.TreeMap;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.RejectedExecutionException;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.BoundedExecutor;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.FileLister;

/**
 * Creates a list of unique tokens extracted from a collection of documents that can be trimmed by removing tokens based
 * on various different attributes; used for creating fingerprints of documents that are based on words that appear in a
 * document collection
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 **/
public class Dictionary implements Serializable {

  /**
   * Comparator that orders strings alphabetically ignoring case; sets the locale to English, Arabic, or Chinese
   **/
  private static class IgnoreCase implements Comparator<String>, Serializable {
    private static final long serialVersionUID = 5081186068508241364L;

    public int compare(String one, String two) {
      return one.compareToIgnoreCase(two);
    }
  }

  // arbitrary serialVersionUID (since we implement Serializable)
  public static final long serialVersionUID = 42L;

  private String creatingProgram;
  private String creation;
  protected Map<String, DictionaryEntry> dictionary;
  private String dictionaryFileName;
  private String dictionaryName;
  private String directory;
  private Map<String, DictionaryEntry> file_input;
  private String GUID;
  private String language;
  private double maxIDFfound;
  private int maxThread;
  protected int nextPosition;
  private Random random;
  private String systemID;
  private boolean threaded;
  private List<String> tokenizerNames;
  private TokenizerList tokenizers;
  protected int totalDocuments;
  private boolean trained;
  private boolean training;
  private boolean trimmed;
  private double trimmed_max;
  private double trimmed_min;
  private int version;

  /**
   * Constructor that sets default values for this dictionary. The default language is English.
   **/
  public Dictionary() {
    language = "english";
    dictionaryName = "";
    GUID = new BigInteger(128, new Random()).toString();
    creatingProgram = setCreatingProgram();
    version = 0;
    threaded = false;
    maxThread = Runtime.getRuntime().availableProcessors();
    random = new Random();
    trimmed = false;
    nextPosition = 0;
    totalDocuments = 0;
    trained = false;
    maxIDFfound = 0.0;
    tokenizerNames = new Vector<String>();
    try {
      systemID = java.net.InetAddress.getLocalHost().getCanonicalHostName();
    } catch (UnknownHostException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Constructor that loads this dictionary from a file
   * 
   * @param dictionaryFilename
   *          the filename containing a dictionary
   */
  public Dictionary(String dictionaryFilename) {
    random = new Random();
    this.loadDictionaryXML(dictionaryFilename);
  }

  /**
   * Adds a document to this dictionary. Requires this dictionary to be in training mode.
   * 
   * @param documentIterator
   *          the iterator over all tokens in the document
   **/
  public void addDocument(Iterator<String> documentIterator) {

    // make sure we are in training mode
    if (training) {

      // iterate over all terms and insert them into a temp hash map
      file_input = new TreeMap<String, DictionaryEntry>(new IgnoreCase());

      while (documentIterator.hasNext()) {

        // get a token
        String token = documentIterator.next();

        // see if it is already in the temp hash map
        if (file_input.containsKey(token)) {
          // if so, increment the frequency count
          file_input.get(token).incrementFrequencyCount();
        } else {
          // otherwise, create a new DictionaryEntry object for the token and
          // insert it
          DictionaryEntry tobeinserted = new DictionaryEntry(token);
          file_input.put(token, tobeinserted);
        }
      }

      mergeDictionary(file_input);

      // clear the temp hash_map
      file_input = null;
    } else {
      ExceptionHandler.handleException(new Exception("Not in training mode"));
    }
  }

  /**
   * Reads an XML file that specifies which tokenizers to use for this dictionary
   * 
   * @param config_file
   *          the string name of the XML configuration file containing this dictionary's tokenizers
   **/
  @SuppressWarnings("unchecked")
  public void addTokenizers(String config_file) {
    try {
      File xml = new File(config_file);
      SAXBuilder builder = new SAXBuilder();
      Document doc = (Document) builder.build(xml);

      Element dict = doc.getRootElement();
      setLanguage(dict.getAttributeValue("language"));
      for (Element tokenizerElement : (List<Element>) dict.getChildren("tokenizer")) {
        String tokenizer = tokenizerElement.getAttributeValue("name");
        if (tokenizerNames.size() == 0 && !tokenizer.contains("FileTokenizer")) {
          ExceptionHandler.handleException(new Exception("Dictionary: First tokenizer (" + tokenizer
              + ") must read from a file."));
        }

        if (tokenizerElement.getAttribute("parameter") != null) {
          tokenizer += " " + tokenizerElement.getAttributeValue("parameter");
        }
        tokenizerNames.add(tokenizer.trim());
      }
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleException(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleException(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleException(e);
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Creates a deep copy of the given dictionary object and stores it in this dictionary
   * 
   * @param toClone
   *          the dictionary object to copy
   */
  public void clone(Dictionary toClone) {
    dictionary.clear();
    dictionary.putAll(toClone.dictionary);
    directory = toClone.directory;
    language = toClone.language;
    nextPosition = toClone.nextPosition;
    trained = toClone.trained;
    training = toClone.training;
    version = toClone.version;
    trimmed = toClone.trimmed;
    if (trimmed) {
      trimmed_min = toClone.trimmed_min;
      trimmed_max = toClone.trimmed_max;
    }
    totalDocuments = toClone.totalDocuments;
    tokenizerNames = toClone.tokenizerNames;
    maxIDFfound = toClone.maxIDFfound;
    System.out.println("Dictionary Clone: directory = " + directory);
  }

  /**
   * Determines if this dictionary contains a particular token
   * 
   * @param token
   *          the string containing the token value
   * 
   * @return the <code>boolean</code> value of whether the token is present in this dictionary
   **/
  public boolean containsToken(String token) {
    return dictionary.containsKey(token);
  }

  /**
   * Creates this dictionary using an optional number or percent of the files in the given fileLister; if no count or
   * percent is specified then uses entire fileList to create dictionary
   * 
   * @param fileList
   *          the fileLister containing the files to use for this dictionary
   */
  public void createFromFileLister(FileLister fileList) {
    dictionary.clear();
    directory = fileList.getDirectory();
    this.createFromIterator(fileList.iterator());
  }

  /**
   * Creates this dictionary using an optional number or percent of the files in the given fileLister
   * 
   * @param fileList
   *          the fileLister containing the files to use for this dictionary
   * @param parameter
   *          the <code>double</code> count (>0) or percent (<0) of the files in the fileLister to use to create this
   *          dictionary
   * 
   * @return the fileLister containing all the files NOT used to create this dictionary
   */
  public FileLister createFromFileLister(FileLister fileList, double parameter) {
    dictionary.clear();
    directory = fileList.getDirectory();

    if (parameter < 1) {
      this.createFromIterator(fileList.createNewListFromPercent(parameter * 10));
    } else {
      this.createFromIterator(fileList.createNewListFromCount((int) parameter));
    }

    return fileList;
  }

  /**
   * Creates this dictionary from the files contained in the given iterator
   * 
   * @param fileIterator
   *          the iterator over the files to use to create this dictionary
   **/
  private void createFromIterator(Iterator<File> fileIterator) {
    if (threaded) {
      createFromIteratorThreaded(fileIterator);
    }

    this.startTraining(tokenizerNames);
    File currentFile;
    // Iterate over the files
    while (fileIterator.hasNext()) {
      currentFile = fileIterator.next();
      if (!currentFile.getName().endsWith(".gif") && !currentFile.getName().endsWith(".jpg")
          && !currentFile.getName().endsWith(".png") && !currentFile.getName().endsWith(".dwf")
          && !currentFile.getName().endsWith(".swf")) {
        // Tokenize the file
        Iterator<String> tokens = tokenizers.tokenizeFile(currentFile);
        if (tokens != null)
          this.addDocument(tokens);
      }
    }
    this.endTraining();
  }

  /**
   * Uses multiple threads to speed execution of creating this dictionary from the files contained in the given iterator
   * 
   * @param fileIterator
   *          the iterator over the files to use to create this dictionary
   **/
  private void createFromIteratorThreaded(Iterator<File> fileIterator) {

    // Put this dictionary into training mode
    try {
      startTraining(tokenizerNames);

      ExecutorService exec = Executors.newFixedThreadPool(maxThread);
      BoundedExecutor bex = new BoundedExecutor(exec, maxThread);

      File currentFile;
      while (fileIterator.hasNext()) {
        currentFile = fileIterator.next();
        if (!currentFile.getName().endsWith(".gif") && !currentFile.getName().endsWith(".jpg")
            && !currentFile.getName().endsWith(".png") && !currentFile.getName().endsWith(".dwf")
            && !currentFile.getName().endsWith(".swf")) {

          DictionaryWorker currentFileProcessor = new DictionaryWorker(currentFile, tokenizerNames, this);
          bex.submitTask(currentFileProcessor);
        }
      }

      exec.shutdown();

      while (!exec.isTerminated()) {
        // Wait a second and see if we are done
        Thread.sleep(1000);
      }
      endTraining();
    } catch (RejectedExecutionException e) {
      ExceptionHandler.handleException(e);
    } catch (InterruptedException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Allows for operations on the tokens in this dictionary
   * 
   * @return the iterator over the tokens in this dictionary
   **/
  public Iterator<String> dictionaryTokenIterator() {
    return dictionary.keySet().iterator();
  }

  /**
   * Ends training, which prohibits adding new tokens to this dictionary and allows this dictionary to be written to a
   * file. Increments the version number and calculates the largest IDF in this dictionary.
   **/
  public void endTraining() {

    if (training) {
      tokenizers = null;
      version++;
      trained = true;
      training = false;
      file_input = null;
      findMaxIDF();
    } else {
      ExceptionHandler.handleException(new Exception("Not in training mode"));
    }
  }

  /**
   * Determines the largest IDF in this dictionary and stores it in the private maxIDFfound variable
   **/
  private void findMaxIDF() {
    Iterator<String> tokens = dictionaryTokenIterator();
    while (tokens.hasNext()) {
      String currentToken = tokens.next();
      maxIDFfound = (getIDF(currentToken) > maxIDFfound) ? getIDF(currentToken) : maxIDFfound;
    }
  }

  /**
   * Provides the program that created this dictionary
   * 
   * @return the string name of the program that created this dictionary
   **/
  public String getCreatingProgram() {
    return creatingProgram;
  }

  /**
   * Provides the original creation date of this dictionary
   * 
   * @return the string date when this dictionary was created
   **/
  public String getCreation() {
    return creation;
  }

  /**
   * Provides the filename of this dictionary, if this dictionary was loaded from or saved to a file
   * 
   * @return the string dictionary filename
   **/
  public synchronized String getDictionaryFilename() {
    return dictionaryFileName;
  }

  /**
   * Provides the name of this dictionary
   * 
   * @return the string name of this dictionary
   **/
  public String getDictionaryName() {
    return dictionaryName;
  }

  /**
   * Provides the number of times the given token appeared when creating this dictionary
   * 
   * @param token
   *          the string containing the token value
   * 
   * @return the number of times the token appeared; returns -1 if the token is not present in this dictionary
   **/
  public int getFrequency(String token) {

    if (!training) {
      if (dictionary.containsKey(token)) {
        return dictionary.get(token).getFrequencyCount();
      } else {
        ExceptionHandler.handleException(new Exception("Token not in dictionary"));
      }
    } else {
      ExceptionHandler.handleException(new Exception("Dictionary in training mode"));
    }
    return -1;
  }

  /**
   * Provides the unique identifier for this dictionary
   * 
   * @return the string GUID for dictionary
   **/
  public String getGUID() {
    return GUID;
  }

  /**
   * Provides the inverse document frequency (IDF) of the given token in this dictionary
   * 
   * @param token
   *          the string containing the token value
   * 
   * @return the <code>double</code> value of the token's inverse document frequency (IDF); returns 0.0 if the token is
   *         not present in this dictionary
   **/
  public synchronized double getIDF(String token) {

    if (!training) {
      if (dictionary.containsKey(token)) {
        return dictionary.get(token).getIDF(totalDocuments);
      } else {
        ExceptionHandler.handleException(new Exception("Token not in dictionary"));
      }
    } else {
      ExceptionHandler.handleException(new Exception("Dictionary in training mode"));
    }
    return 0.0;
  }

  /**
   * Provides the language of this dictionary
   * 
   * @return the string language of the files used to create this dictionary
   */
  public String getLanguage() {
    return language;
  }

  /**
   * Provides the largest IDF value in this dictionary
   * 
   * @return the <code>double</code> value of the largest IDF in this dictionary
   */
  public double getMaxIDF() {
    return maxIDFfound;
  }

  /**
   * Provides the normalized inverse document frequency (IDF) of the given token in this dictionary. The normalized IDF
   * is the token's IDF divided by the highest IDF in this dictionary.
   * 
   * @param token
   *          the string containing the token value
   * 
   * @return the <code>double</code> value of the token's normalized inverse document frequency (IDF); returns 0.0 if
   *         the token is not present in this dictionary
   **/
  public synchronized double getNormalizedIDF(String token) {
    if (!training) {
      if (dictionary.containsKey(token)) {
        return (dictionary.get(token).getIDF(totalDocuments) / maxIDFfound);
      } else {
        ExceptionHandler.handleException(new Exception("Token not in dictionary"));
      }
    } else {
      ExceptionHandler.handleException(new Exception("Dictionary in training mode"));
    }
    return 0.0;
  }

  /**
   * Provides the position of the given token in this dictionary
   * 
   * @param token
   *          the string containing the token value
   * 
   * @return the position of the token; returns -1 if the token is not present in this dictionary
   **/
  public int getPosition(String token) {

    if (!training) {
      if (dictionary.containsKey(token)) {
        return dictionary.get(token).getPosition();
      } else {
        return -1;
      }

    } else {
      ExceptionHandler.handleException(new Exception("Dictionary in training mode"));
    }
    return -1;
  }

  /**
   * Gives a random token from this dictionary
   * 
   * @return the string value of the random token from this dictionary
   **/
  public String getRandomToken() {

    if (null == dictionary) {
      ExceptionHandler.handleException(new Exception("Dictionary: Cannot return random token from null dictionary"));
    }

    List<String> tokens = new Vector<String>(dictionary.keySet());

    if (dictionary.size() != tokens.size()) {
      ExceptionHandler.handleException(new Exception("Dictionary: Keys appeared or disappeared"));
    }
    if (random == null) {
      ExceptionHandler.handleException(new Exception("Dictionary: Null Random Number Generator"));
    }
    return tokens.get(random.nextInt(tokens.size()));
  }

  /**
   * Provides the number of tokens in this dictionary
   * 
   * @return the number of unique tokens in this dictionary
   **/
  public int getSize() {
    return dictionary.size();
  }

  /**
   * Provides the directory used for this dictionary
   * 
   * @return the string directory for which this dictionary was created
   **/
  public String getSource() {
    return directory;
  }

  /**
   * Provides the identifier for the system that this dictionary was created on
   * 
   * @return the string system ID for this dictionary
   **/
  public String getSystemID() {
    return systemID;
  }

  /**
   * Provides the names of the tokenizers used to create this dictionary
   * 
   * @return the string list of all the tokenizer names used for this dictionary
   **/
  public List<String> getTokenizerNames() {
    return tokenizerNames;
  }

  /**
   * Provides a vector of the tokens in this dictionary
   * 
   * @return the list of tokens in this dictionary
   **/
  public List<String> getTokens() {
    Iterator<String> tokensIterator = dictionaryTokenIterator();
    List<String> tokens = new Vector<String>();
    while (tokensIterator.hasNext()) {
      tokens.add(tokensIterator.next());
    }
    return tokens;
  }

  /**
   * Provides the number of documents processed for this dictionary
   * 
   * @return the total number of documents used to create this dictionary
   **/
  public synchronized int getTotalDocuments() {
    return totalDocuments;
  }

  /**
   * Provides the version number of this dictionary. The version is incremented when training on this dictionary ends.
   * 
   * @return the version number of this dictionary
   **/
  public int getVersion() {
    return version;
  }

  /**
   * Loads a previously serialized dictionary object into this dictionary
   * 
   * @param filename
   *          the string filename of the dictionary file to load
   **/
  public synchronized void loadDictionary(String filename) {

    try {
      FileInputStream fileIn = new FileInputStream(filename);
      ObjectInputStream in = new ObjectInputStream(fileIn);
      Dictionary incoming = (Dictionary) in.readObject();
      this.clone(incoming);
      dictionaryFileName = filename;
      in.close();
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleException(e);
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    } catch (ClassNotFoundException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Loads a previously serialized XML dictionary object into this dictionary
   * 
   * @param filename
   *          the string filename of the XML dictionary file to load
   **/
  @SuppressWarnings("unchecked")
  public synchronized void loadDictionaryXML(String filename) {

    if (dictionary != null)
      dictionary.clear();
    tokenizerNames = new Vector<String>();
    dictionaryFileName = filename;

    try {
      File xml = new File(dictionaryFileName);
      SAXBuilder builder = new SAXBuilder();
      Document doc = (Document) builder.build(xml);
      Element dict = doc.getRootElement();

      dictionaryName = dict.getAttributeValue("name");
      GUID = dict.getAttributeValue("GUID");
      systemID = dict.getAttributeValue("systemID");
      version = Integer.valueOf(dict.getAttributeValue("version"));
      setLanguage(dict.getAttributeValue("language"));
      creatingProgram = dict.getAttributeValue("creatingProgram");
      creation = dict.getAttributeValue("creation");
      directory = dict.getAttributeValue("directory");
      maxIDFfound = Double.valueOf(dict.getAttributeValue("maxIDF"));
      totalDocuments = Integer.valueOf(dict.getAttributeValue("totalDocuments"));
      trimmed = Boolean.valueOf(dict.getAttributeValue("trimmed"));
      if (trimmed) {
        trimmed_min = Double.valueOf(dict.getAttributeValue("trimmed_min"));
        trimmed_max = Double.valueOf(dict.getAttributeValue("trimmed_max"));
      }
      trained = Boolean.valueOf(dict.getAttributeValue("trained"));
      training = Boolean.valueOf(dict.getAttributeValue("training"));
      nextPosition = Integer.valueOf(dict.getAttributeValue("nextPosition"));

      for (Element tokenizerElement : (List<Element>) dict.getChildren("tokenizer")) {
        String tokenizer = tokenizerElement.getAttributeValue("name");
        if (tokenizerElement.getAttribute("parameter") != null) {
          tokenizer += " " + tokenizerElement.getAttributeValue("parameter");
        }
        tokenizerNames.add(tokenizer.trim());
      }

      for (Element tokenElement : (List<Element>) dict.getChildren("token")) {
        dictionary.put(
            tokenElement.getAttributeValue("name"),
            new DictionaryEntry(tokenElement.getAttributeValue("name"), Integer.valueOf(tokenElement
                .getAttributeValue("position")), Integer.valueOf(tokenElement.getAttributeValue("frequency")), Integer
                .valueOf(tokenElement.getAttributeValue("documentCount")), Double.valueOf(tokenElement
                .getAttributeValue("idf"))));
      }
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleException(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleException(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleException(e);
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Creates dictionary from the given directory; Reads an XML file that specifies which tokenizers to use and outputs
   * dictionary to a file
   * 
   * @param config_file
   *          the string name of the XML configuration file containing dictionary tokenizers
   * @param directory
   *          the string path to the directory used to create this dictionary
   * @param output_file
   *          the string name/path to write this dictionary to
   */
  public void makeDictionary(String config_file, String directory, String output_file) {

    if (output_file == null || config_file == null || directory == null) {
      ExceptionHandler.handleException(new Exception("Dictionary: Null parameter(s) given."));
    }

    addTokenizers(config_file);

    if (tokenizerNames == null) {
      ExceptionHandler.handleException(new Exception("Dictionary: Null tokenizer vector"));
    }
    if (tokenizerNames.size() < 1) {
      ExceptionHandler.handleException(new Exception("Dictionary: Empty tokenizer vector"));
    }
    if (dictionary == null) {
      ExceptionHandler.handleException(new Exception("Dictionary: Null dictionary"));
    }

    setTokenizersByName(tokenizerNames);

    // Don't thread if OutsideInFileTokenizer is being used
    if (tokenizerNames.get(0).contains("OutsideInFileTokenizer")) {
      threadingOff();
    } else {
      threadingOn();
    }

    FileLister files = new FileLister();
    files.recursiveLoadDirectory(directory);
    createFromFileLister(files);

    saveDictionaryXML(output_file);
  }

  /**
   * Merges the given Map of tokens to dictionaryEntry with this dictionary
   * 
   * @param fileInput
   *          the map containing the dictionary information to merge with this dictionary
   */
  protected synchronized void mergeDictionary(Map<String, DictionaryEntry> fileInput) {
    for (Entry<String, DictionaryEntry> entry : fileInput.entrySet()) {
      if (dictionary.containsKey(entry.getKey())) {
        // If so, increment the document count and frequency count
        dictionary.get(entry.getKey()).incrementFrequencyCount(entry.getValue().getFrequencyCount());
        dictionary.get(entry.getKey()).incrementDocumentCount();
      } else {
        // Otherwise, insert the DictionaryEntry into this dictionary
        entry.getValue().setPosition(nextPosition);
        nextPosition++;
        dictionary.put(entry.getKey(), entry.getValue());
      }
    }
    totalDocuments++;
  }

  /**
   * Writes this dictionary as a serialized object. Requires dictionary to be out of training mode.
   * 
   * @param filename
   *          the string filename to write this dictionary to
   **/
  public void saveDictionary(String filename) {

    if (!training) {

      // Don't need to save the random number generator
      random = null;

      dictionaryFileName = filename;
      try {
        FileOutputStream fout = new FileOutputStream(dictionaryFileName);
        ObjectOutputStream out = new ObjectOutputStream(fout);
        out.writeObject(this);
        out.close();
        fout.close();
      } catch (FileNotFoundException e) {
        ExceptionHandler.handleException(e);
      } catch (IOException e) {
        ExceptionHandler.handleException(e);
      }
    } else {
      ExceptionHandler.handleException(new Exception("No saves in training mode."));
    }
  }

  /**
   * Writes this dictionary as a serialized XML object. Requires dictionary to be out of training mode.
   * 
   * @param filename
   *          the string XML filename to write this dictionary to
   **/
  public void saveDictionaryXML(String filename) {

    if (!training) {
      // Don't need to save the random number generator
      random = null;
      dictionaryFileName = (new File(filename)).getAbsolutePath();
      dictionaryName = filename.substring(filename.lastIndexOf('/') + 1);
      creation = String.valueOf(new Date());
      DecimalFormat rounding = new DecimalFormat("#0.####");

      Element dict = new Element("dictionary");
      dict.setAttribute("name", dictionaryFileName);
      dict.setAttribute("GUID", GUID);
      dict.setAttribute("systemID", systemID);
      dict.setAttribute("version", String.valueOf(version));
      dict.setAttribute("language", language);
      if (getCreatingProgram() != null && getCreatingProgram().length() > 0)
        dict.setAttribute("creatingProgram", creatingProgram);
      if (getCreation() != null && getCreation().length() > 0)
        dict.setAttribute("creation", creation);
      if (getSource() != null && getSource().length() > 0)
        dict.setAttribute("directory", directory);
      dict.setAttribute("maxIDF", rounding.format((maxIDFfound)));
      dict.setAttribute("totalDocuments", String.valueOf(totalDocuments));
      dict.setAttribute("trimmed", String.valueOf(trimmed));
      if (trimmed) {
        dict.setAttribute("trimmed_min", String.valueOf(trimmed_min));
        dict.setAttribute("trimmed_max", String.valueOf(trimmed_max));
      }
      dict.setAttribute("trained", String.valueOf(trained));
      dict.setAttribute("training", String.valueOf(training));
      dict.setAttribute("nextPosition", String.valueOf(nextPosition));

      for (String tokenizer : tokenizerNames) {
        String[] split = tokenizer.split("\\s");
        Element tokenizerElement = new Element("tokenizer");
        tokenizerElement.setAttribute("name", split[0]);
        StringBuffer parameters = new StringBuffer();
        if (split.length > 1) {
          parameters.append(split[1]);
          tokenizerElement.setAttribute("parameter", split[1]);
        }
        dict.addContent(tokenizerElement);
      }

      for (String token : dictionary.keySet()) {
        Element tokenElement = new Element("token");
        try {
          tokenElement.setAttribute("name", token);
          tokenElement.setAttribute("position", String.valueOf(getPosition(token)));
          tokenElement.setAttribute("frequency", String.valueOf(getFrequency(token)));
          tokenElement.setAttribute("documentCount", String.valueOf(dictionary.get(token).getDocumentCount()));
          tokenElement.setAttribute("idf", rounding.format((getIDF(token))));
          tokenElement.setAttribute("idfNorm", rounding.format((getNormalizedIDF(token))));
          dict.addContent(tokenElement);
        } catch (Exception e) {
          System.err.println("token: \"" + token + "\"");
          ExceptionHandler.handleException(e);
        }
      }
      try {
        Document doc = new Document(dict);
        FileOutputStream dictionaryOut = new FileOutputStream(dictionaryFileName);
        Format form = Format.getPrettyFormat();
        form.setEncoding("UTF-8");
        XMLOutputter out = new XMLOutputter(form);
        out.output(doc, dictionaryOut);
        System.out.println("Writing dictionary to " + dictionaryFileName);
        dictionaryOut.close();
      } catch (FileNotFoundException e) {
        ExceptionHandler.handleException(e);
      } catch (IOException e) {
        ExceptionHandler.handleException(e);
      }
    } else {
      ExceptionHandler.handleException(new Exception("No saves in training mode."));
    }
  }

  /**
   * Determines the name of the program that created this dictionary
   * 
   * @return the string value of the program used to create this dictionary, returns null if no program can be
   *         determined
   **/
  public String setCreatingProgram() {
    creatingProgram = "";
    Runtime runtime = Runtime.getRuntime();
    String cmds[] = { "jps", "-m" };
    Process proc;
    try {
      proc = runtime.exec(cmds);
      InputStreamReader in = new InputStreamReader(proc.getInputStream(), "UTF-8");
      BufferedReader bufferedreader = new BufferedReader(in);

      String line;
      while ((line = bufferedreader.readLine()) != null) {
        for (String output : line.split("\\s")) {
          if (output.contains(".jar")) {
            in.close();
            bufferedreader.close();
            return output;
          }
        }
      }
      in.close();
      bufferedreader.close();
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
    return null;
  }

  /**
   * Sets the name of the program that created this dictionary
   * 
   * @param creatingProgram
   *          the string program that created this dictionary
   */
  public void setCreatingProgram(String creatingProgram) {
    this.creatingProgram = creatingProgram;
  }

  /**
   * Sets the language of this dictionary
   * 
   * @param language
   *          the string language of the files used to create this dictionary
   */
  public void setLanguage(String language) {
    this.language = (language != null) ? language : "english";
    dictionary = new TreeMap<String, DictionaryEntry>(new IgnoreCase());
  }

  /**
   * Sets a random number generator to allow for repeatability. Repeatable behavior is sometimes used when running
   * experiments.
   * 
   * @param random
   *          the random object to use for this dictionary
   **/
  public void setRandomNumberGenerator(Random random) {
    this.random = random;
  }

  /**
   * Loads the tokenizers to use for this dictionary from a TokenizerList object. Only works if no tokenizers have
   * already been set.
   * 
   * @param tokenizerList
   *          the TokenizerList object containing the tokenizers to use for this dictionary
   */
  public void setTokenizers(TokenizerList tokenizerList) {

    if (tokenizers == null) {
      tokenizerNames = tokenizerList.getNames();
    } else {
      ExceptionHandler.handleException(new Exception("Dictionary: Cannot reset tokenizer list in a dictionary"));
    }
  }

  /**
   * Sets the tokenizers used by this dictionary; Only works if no tokenizers have already been set or if the given list
   * contains the same tokenizers as those that have already been set
   * 
   * @param tokenizerVec
   *          the list of tokenizer names to use to create this dictionary
   **/
  public void setTokenizersByName(List<String> tokenizerVec) {

    if (null != tokenizerVec) {
      if (null == tokenizerNames) {
        tokenizerNames = tokenizerVec;
      } else {
        if ((tokenizerVec.size() != tokenizerNames.size()) || (!tokenizerVec.containsAll(tokenizerNames)))
          ExceptionHandler.handleException(new Exception("Dictionary: Cannot reset tokenizers in a dictionary"));
        else {
          tokenizers = new TokenizerList(tokenizerNames);
        }
      }
    } else {
      ExceptionHandler.handleException(new Exception("Dictionary: Tokenizer name list is null"));
    }
  }

  /**
   * Prints information about this dictionary, including all tokens and their statistics.
   **/
  public synchronized void showDictionary() {

    System.out.println("Dictionary File: " + dictionaryFileName);
    System.out.println("Language: " + language);
    System.out.println("Version: " + version);
    System.out.println("Total Documents: " + totalDocuments);
    System.out.println("---------------------------------");
    Iterator<String> keys = dictionaryTokenIterator();
    while (keys.hasNext()) {
      String key = keys.next();
      System.out.println(" " + key);
      System.out.println("Pos:    " + getPosition(key));
      System.out.println("Fre:    " + getFrequency(key));
      System.out.println("IDF:    " + getIDF(key));
      System.out.println();
    }
  }

  /**
   * Prints this dictionary statistics, including:
   * <ul>
   * <li>Name</li>
   * <li>Language</li>
   * <li>Number of documents</li>
   * <li>Number of tokens</li>
   * <li>If this dictionary has been trimmed (including the IDF range, if trimmed)</li>
   * <li>Minimum and maximum IDF</li>
   * <li>List of tokenizers</li>
   * </ul>
   **/
  public void showStatistics() {

    System.out.println("Name: " + dictionaryFileName);
    System.out.println("Language: " + language);
    System.out.println("# Docs: " + totalDocuments);
    System.out.println("# Tokens: " + dictionary.size());
    if (trimmed) {
      System.out.println("Trimmed: "+trimmed+" (min: " + trimmed_min + ", max: " + trimmed_max+")");
    } else {
      System.out.println("Trimmed: " + trimmed);
    }
    System.out.println("Max IDF: " + maxIDFfound);
    System.out.println("Tokenizers:");
    showTokenizers();
  }

  /**
   * Prints the names of the tokenizers used for this dictionary
   **/
  public void showTokenizers() {
    for (String tokenizer : tokenizerNames) {
      System.out.println("    " + tokenizer);
    }
  }

  /**
   * Prints each token in this dictionary with it's frequency, IDF, and normalized IDF
   **/
  public void showTokens() {
    Iterator<String> tokens = dictionaryTokenIterator();
    System.out.format("%-25s %-10s %-10s %-10s%n", "Token", "Freq", "IDF", "Norm. IDF");
    System.out.println("----------------------------------------------------------");
    while (tokens.hasNext()) {
      String token = tokens.next();
      System.out.format("%-25s %-10s %-10.4f %-10.4f%n", token, getFrequency(token), getIDF(token),
          getNormalizedIDF(token));
    }
  }

  /**
   * Starts training, which permits tokens to be added to this dictionary. Specifies the names of the tokenizers to use
   * on the new tokens.
   * 
   * @param tokenizerVec
   *          the string list containing tokenizer names; must match the tokenizers used to create this dictionary
   **/
  public void startTraining(List<String> tokenizerVec) {

    if (trimmed) {
      ExceptionHandler.handleException(new Exception("Dictionary: Training a trimmed dictionary is a bad idea"));
    }

    // The tokenizerNames variable holds the names of the tokenizers used to
    // create this dictionary. We need to make sure they are consistent
    if (tokenizers == null) {
      if (tokenizerNames == null) {
        tokenizerNames = tokenizerVec;
      } else if (tokenizerNames.size() != tokenizerVec.size() || !tokenizerVec.containsAll(tokenizerNames)) {
        ExceptionHandler
            .handleException(new Exception("Dictionary: Must use same tokenizers for adding to Dictionary"));
      }
      setTokenizersByName(tokenizerNames);
    }
    training = true;
  }

  /**
   * Turns off threading for creating this dictionary
   **/
  public void threadingOff() {
    threaded = false;
  }

  /**
   * Turns on threading for creating this dictionary. This should only be used for stand-alone dictionary creation (not
   * inside a threaded application).
   **/
  public void threadingOn() {
    threaded = true;
  }

  /**
   * Turns on threading for creating this dictionary. This should only be used for stand-alone dictionary creation (not
   * inside a threaded application).
   * 
   * @param numThreads
   *          the number of threads to use to create this dictionary
   **/
  public void threadingOn(int numThreads) {
    maxThread = numThreads;
    threaded = true;
  }

  /**
   * Trims this dictionary by removing any token that is outside a range of normalized IDFs. Once trimmed, no new tokens
   * can be added to this dictionary.
   * 
   * @param min
   *          the <code>double</code> minimum normalized IDF to keep; any token with a lower IDF will be discarded.
   * @param max
   *          the <code>double</code> maximum normalized IDF to keep; any token with a higher IDF will be discarded.
   **/
  public void trimByIDF(double min, double max) {

    // Trims this dictionary down to contain only tokens within the particular
    // range of normalized IDF values
    if (trimmed) {
      ExceptionHandler.handleException(new Exception(
          "Dictionary: trimming an already trimmed dictionary is probably a bad idea"));
    }

    trimmed = true;
    trimmed_min = min;
    trimmed_max = max;
    int currentPlace = 0;

    Map<String, DictionaryEntry> trimmedDictionary = new TreeMap<String, DictionaryEntry>(new IgnoreCase());

    Iterator<String> tokens = dictionaryTokenIterator();
    while (tokens.hasNext()) {
      String token = tokens.next();
      if ((getNormalizedIDF(token) >= min) && (getNormalizedIDF(token) <= max)) {
        DictionaryEntry entry = dictionary.get(token);
        entry.setPosition(currentPlace++);
        trimmedDictionary.put(token, entry);
      }
    }

    dictionary = trimmedDictionary;
  }

  /**
   * Creates a dictionary from the specified directory using the provided tokenizers
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename of the tokenizer configuration file<br>
   *          <code>args[1]</code> the directory to create the dictionary from<br>
   *          <code>args[2]</code> the filename to write the dictionary to
   */
  public static void main(String args[]) {
    String output_file = null;
    String directory = null;
    String config_file = null;

    try {
      config_file = args[0];
      directory = args[1];
      output_file = args[2];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err
          .println("Usage: java CreateDictionary <tokenizers config file> <directory> <output dictionary filename> ");
      System.err.println("Recursively creates a dictionary from the specified directory using the provided tokenizers");
      System.exit(-1);
    }
    Dictionary dict = new Dictionary();
    dict.makeDictionary(config_file, directory, output_file);
    dict.showTokens();
  }
}
