package edu.georgetown.gucs.dictionary;

/**
 * Creates a {@linkplain Dictionary} containing tokens with IDFs within a specified range
 * 
 * @author Clay Shields
 */
public class TrimDictionary {

  private Dictionary dictionary;
  private double min_idf = 0.0;
  private double max_idf = 1.0;

  /**
   * Constructor that specifies the IDF range and dictionary to trim
   * 
   * @param min_idf
   *          the <code>double</code> minimum normalized inverse document frequency (IDF) to use for this trimmed
   *          dictionary
   * @param max_idf
   *          the <code>double</code> maximum normalized inverse document frequency (IDF) to use for this trimmed
   *          dictionary
   * @param dictionary_file
   *          the string filename containing the dictionary to trim
   **/
  public TrimDictionary(double min_idf, double max_idf, String dictionary_file) {
    dictionary = new Dictionary(dictionary_file);
    this.min_idf = min_idf;
    this.max_idf = max_idf;
  }

  /**
   * 
   * @param output_file
   *          the string filename to write the trimmed dictionary to
   */
  public void trim(String output_file) {
    dictionary.trimByIDF(min_idf, max_idf);
    dictionary.saveDictionaryXML(output_file);
  }

  /**
   * Trims a dictionary by removing any token that is outside a range of normalized IDFs
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the minimum normalized inverse document frequency (IDF) to use for this trimmed
   *          dictionary<br>
   *          <code>args[1]</code> the maximum normalized inverse document frequency (IDF) to use for this trimmed
   *          dictionary<br>
   *          <code>args[2]</code> the filename containing the dictionary to trim<br>
   *          <code>args[3]</code> the optional filename to write the trimmed dictionary to
   */
  public static void main(String args[]) {
    double min_idf = 0.0;
    double max_idf = 1.0;
    String dictionary_file = null;
    String output_file = null;
    try {
      min_idf = Double.valueOf(args[0]);
      max_idf = Double.valueOf(args[1]);
      dictionary_file = args[2];
      if (args.length > 3)
        output_file = args[3];
      else
        output_file = dictionary_file.substring(0, dictionary_file.indexOf(".")) + "_trimmed_" + min_idf + "-"
            + max_idf + ".xml";
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err
          .println("Usage: java TrimDictionary <min IDF> <max IDF> <input file name> <optional: output file name>");
      System.exit(-1);
    }

    TrimDictionary trimDictionary = new TrimDictionary(min_idf, max_idf, dictionary_file);
    trimDictionary.trim(output_file);
  }
}
