package edu.georgetown.gucs.experiment;

import java.io.File;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.matcher.ExactFingerprintMatcher;
import edu.georgetown.gucs.matcher.FingerprintMatcher;
import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.ObjectBuilder;

/**
 * Uses a given set of parameters to create a dictionary and fingerprints of a set of documents. Matches are based on
 * comparisons of unmangled and mangled versions of each file and optionally stores the matches in a database.
 * 
 * @author Lindsay Neubauer
 * @author Clay Shields
 * 
 */
public class Trial implements Runnable {

  private TrialParameters parameters;
  private int experimentID;
  private boolean insert;
  private DBInterface database;
  private FileLister datasetFiles;
  private File datasetPath;
  private String language;
  private Random random;
  private long randomSeed;
  private double dictionaryParameter;
  private double sampleParameter;
  private double minIDF;
  private double maxIDF;
  private TokenizerList tokenizers;
  private String manglers;
  private Fingerprinter fingerprinter;
  private FingerprintMatcher matcher;
  private Dictionary dictionary;
  private int trialID;
  private int dictionaryID;
  private boolean finished = false;

  private int manglerIDCount = 0;
  private int fileIDCount = 0;
  private int fingerprintIDCount = 0;

  private Map<Integer, byte[]> fingerprintID_byte = new HashMap<Integer, byte[]>();
  private Map<byte[], Integer> fingerprintByte_id = new HashMap<byte[], Integer>();

  private Map<Integer, Integer> fingerprintID_manglerID = new HashMap<Integer, Integer>();
  private Map<Integer, Integer> fingerprintID_fileID = new HashMap<Integer, Integer>();

  private Map<Integer, String> manglerID_settings = new HashMap<Integer, String>();
  private Map<String, Integer> manglerSettings_id = new HashMap<String, Integer>();

  private Map<Integer, File> fileID_path = new HashMap<Integer, File>();
  private Map<File, Integer> filePath_id = new HashMap<File, Integer>();

  private Set<Integer> unmangledFingerprints = Collections.synchronizedSet(new HashSet<Integer>());
  private Set<Integer> mangledFingerprints = Collections.synchronizedSet(new HashSet<Integer>());

  private Map<String, Double[]> manglerResults = new LinkedHashMap<String, Double[]>();

  /**
   * Constructor that sets default values for this trial
   * 
   * @param experimentID
   *          the <code>int</code> value of the experiment id this trial is a part of; set to -1 if insert = false
   * @param insert
   *          the <code>boolean</code> value of whether to insert the results from this trial into the database
   * @param database
   *          the dBInterface database to insert trial results; null if insert = false
   * @param dataset
   *          the string name of the database to insert trial results
   * @param datasetFiles
   *          the fileLister containing the files to use for the dictionary and sample in this trial
   * @param datasetPath
   *          the path to the files to use for the dictionary and sample in this trial
   * @param language
   *          the string language of the files used in this trial
   * @param randomSeed
   *          the <code>long</code> the initial seed to use for the random number generator in this trial; can be used
   *          to enable repeatability across trials
   * @param dictionaryParameter
   *          the <code>double</code> containing the count or percent to use for dictionary creation in this trial;
   *          numbers less than 1 are processed as percents, numbers greater than or equal to 1 are processed as counts
   * @param sampleParameter
   *          the <code>double</code> containing the count or percent to use for the file sample size in this trial;
   *          numbers less than 1 are processed as percents, numbers greater than or equal to 1 are processed as counts
   * @param minIDF
   *          the <code>double</code> minimum normalized IDF to keep in the dictionary; any token with a lower IDF will
   *          be discarded
   * @param maxIDF
   *          the <code>double</code> maximum normalized IDF to keep in the dictionary; any token with a higher IDF will
   *          be discarded
   * @param tokenizers
   *          the list of tokenizer strings to use in this trial
   * @param manglers
   *          the list different mangler strings to use in this trial
   * @param fingerprinter
   *          the string name of the fingerprinter to use in this trial
   * @param matcher
   *          the string name of the matcher to use for fingerprint comparisons in this trial
   * @param matcherScore
   *          the <code>int</code> minimum score that the matcher considers a match
   */
  public Trial(int experimentID, boolean insert, DBInterface database, String dataset, FileLister datasetFiles,
      File datasetPath, String language, long randomSeed, double dictionaryParameter, double sampleParameter,
      double minIDF, double maxIDF, List<String> tokenizers, String manglers, String fingerprinter, String matcher,
      int matcherScore) {

    this.experimentID = experimentID;
    this.insert = insert;
    this.database = database;
    this.datasetFiles = datasetFiles;
    this.datasetPath = datasetPath;
    this.language = language;
    this.randomSeed = randomSeed;
    random = new Random(randomSeed);
    this.dictionaryParameter = dictionaryParameter;
    this.sampleParameter = sampleParameter;
    this.minIDF = minIDF;
    this.maxIDF = maxIDF;
    this.tokenizers = new TokenizerList(tokenizers);
    this.fingerprinter = (Fingerprinter) ObjectBuilder.buildObject(fingerprinter);
    this.matcher = (FingerprintMatcher) ObjectBuilder.buildObject(matcher);
    this.matcher.setMinimumScore(matcherScore);
    this.manglers = manglers;

    this.parameters = new TrialParameters(dataset, Double.toString(sampleParameter),
        Double.toString(dictionaryParameter), Double.toString(minIDF), Double.toString(maxIDF), tokenizers.toString(),
        manglers, this.fingerprinter.getClass().getSimpleName(), this.matcher.getClass().getSimpleName(),
        Integer.toString(matcherScore));
  }

  /**
   * Provides the parameters for this trial that are needed for results comparisons
   * 
   * @return the trialParameters containing the relevant parameters
   */
  public TrialParameters getTrialParameters() {
    return parameters;
  }

  /**
   * Provides the results for each mangler; sleeps until the trial is finished
   * 
   * @return the map containing the manglers results from this trial
   */
  public Map<String, Double[]> getManglerResults() {
    while (!finished) {
      try {
        Thread.sleep(1000);
      } catch (InterruptedException e) {
        ExceptionHandler.handleException(e);
      }
    }
    return manglerResults;
  }

  /**
   * Provides the size of the dictionary used
   * 
   * @return the <code>int</code> size of the dictionary in this trial
   */
  public int getDictionarySize() {
    return dictionary.getSize();
  }

  /**
   * Runs this trial
   * 
   * @return the <code>int</code> trialID from this trial
   */
  public int runTrial() {
    run();
    return trialID;
  }

  /**
   * Runs the trial
   */
  public void run() {

    if (insert) {
      trialID = database.newTrial(randomSeed);
      database.addTrialToExperiment(experimentID, trialID);
    }

    if (datasetFiles == null) {
      datasetFiles = new FileLister();
      datasetFiles.recursiveLoadDirectory(datasetPath.getAbsolutePath());
    }

    FileLister remainingFiles = createDictionary();
    this.fingerprinter.setDictionary(dictionary);
    this.fingerprinter.setManglerRNG(random);
    this.fingerprinter.setMangler(manglers, dictionary);

    if (sampleParameter < 1) {
      remainingFiles.createNewListFromPercent(sampleParameter * 10);
    } else {
      remainingFiles.createNewListFromCount((int) sampleParameter);
    }

    computeFingerprints(null, remainingFiles.getNewListIterator());
    computeFingerprints(manglers, remainingFiles.getNewListIterator());

    doComparisons();
    calculateTrialResults();
  }

  /**
   * Creates a dictionary using a set number or percentage of the files in the dataset fileLister and optionally stores
   * the dictionary in the database
   * 
   * @return the fileLister containing the files from the directory that were NOT used in dictionary creation
   */
  private FileLister createDictionary() {
    dictionary = new Dictionary();
    dictionary.setLanguage(language);
    dictionary.setRandomNumberGenerator(random);
    dictionary.setTokenizers(tokenizers);

    FileLister remainingFiles = dictionary.createFromFileLister(datasetFiles, dictionaryParameter);

    // trim dictionary if needed before writing
    if ((minIDF > 0) || (maxIDF < 1)) {
      dictionary.trimByIDF(minIDF, maxIDF);
    }

    if (insert) {
      dictionaryID = database.newDictionary(dictionary);
      database.addDictionaryToTrial(trialID, dictionaryID);
    }
    return remainingFiles;
  }

  /**
   * Computes fingerprints for each of the given files
   * 
   * @param manglerSettings
   *          the string containing the mangler settings to use; null if taking unmangled fingerprints
   * @param files
   *          the file iterator for the files to fingerprint
   */
  private void computeFingerprints(String manglerSettings, Iterator<File> files) {
    boolean unmangled = false;
    if (manglerSettings == null) {
      unmangled = true;
      manglerSettings = "null";
    }

    int currentManglerID, currentFileID, currentFingerprintID;
    while (files.hasNext()) {
      File currentFile = files.next();
      byte[] fingerprint = fingerprinter.computeFingerprint(currentFile.toString());

      if (fileID_path.containsValue(currentFile)) {
        currentFileID = filePath_id.get(currentFile);
      } else {
        fileIDCount++;
        currentFileID = fileIDCount;
        fileID_path.put(currentFileID, currentFile);
        filePath_id.put(currentFile, currentFileID);
      }

      if (insert) {
        currentManglerID = database.getManglerID(manglerSettings);
        if (manglerIDCount < currentManglerID)
          manglerIDCount = currentManglerID;
        database.addManglerToTrial(trialID, currentManglerID);
        database.insertFingerprint(currentFile, fingerprint, trialID, currentManglerID);
      } else if (manglerID_settings.containsValue(manglerSettings)) {
        currentManglerID = manglerSettings_id.get(manglerSettings);
        if (manglerIDCount < currentManglerID)
          manglerIDCount = currentManglerID;
      } else {
        manglerIDCount++;
        currentManglerID = manglerIDCount;
      }
      manglerID_settings.put(currentManglerID, manglerSettings);
      manglerSettings_id.put(manglerSettings, currentManglerID);

      if (fingerprintID_byte.containsValue(fingerprint)) {
        currentFingerprintID = fingerprintByte_id.get(fingerprint);
      } else {
        fingerprintIDCount++;
        currentFingerprintID = fingerprintIDCount;
        fingerprintID_byte.put(currentFingerprintID, fingerprint);
        fingerprintID_manglerID.put(currentFingerprintID, currentManglerID);
        fingerprintID_fileID.put(currentFingerprintID, currentFileID);
      }

      if (unmangled) {
        unmangledFingerprints.add(currentFingerprintID);
      } else {
        mangledFingerprints.add(currentFingerprintID);
      }
    }
  }

  /**
   * Compares each mangled and unmangled fingerprint using this trials matcher and an exactFingerprintMatcher and
   * records true-positive, false-positive, true-negative, and false-negative results based on this comparison
   */
  private void doComparisons() {
    ExactFingerprintMatcher exactMatcher = new ExactFingerprintMatcher();

    for (Integer mangledFingerprintID : mangledFingerprints) {
      byte[] mangledFP = fingerprintID_byte.get(mangledFingerprintID);
      String mangledFile = fileID_path.get(fingerprintID_fileID.get(mangledFingerprintID)).getPath();
      byte[] mangledFPWithoutMangling = fingerprinter.computeFingerprint(mangledFile);

      for (Integer unmangledFingerprintID : unmangledFingerprints) {
        byte[] unmangledFP = fingerprintID_byte.get(unmangledFingerprintID);

        // Match and compare to an ExactFingerprint match of the unmangled
        // fingerprints from each file
        boolean matchClaim = matcher.match(unmangledFP, mangledFP);
        boolean exactMatch = exactMatcher.match(unmangledFP, mangledFPWithoutMangling);

        // Set up results
        String manglers = manglerID_settings.get(fingerprintID_manglerID.get(mangledFingerprintID));
        Double[] result = manglerResults.get(manglers);
        if (result == null) {
          result = new Double[] { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
        }

        // Update results
        if (matchClaim == true && exactMatch == true) {
          result[Global.TRUE_POSITIVE]++;
        } else if (matchClaim == true && exactMatch == false) {
          result[Global.FALSE_POSITIVE]++;
        } else if (matchClaim == false && exactMatch == true) {
          result[Global.FALSE_NEGATIVE]++;
        } else if (matchClaim == false && exactMatch == false) {
          result[Global.TRUE_NEGATIVE]++;
        }
        manglerResults.put(manglers, result);

        if (insert) {
          int score = matcher.getScore(unmangledFP, mangledFP);
          database.addToFingerprintMatch(trialID, unmangledFingerprintID, mangledFingerprintID, matchClaim, exactMatch,
              score);
        }
      }
    }
  }

  /**
   * Computes the precision, recall, and f-score for each mangler
   */
  public void calculateTrialResults() {

    for (Integer fingerprintID : mangledFingerprints) {
      String manglers = manglerID_settings.get(fingerprintID_manglerID.get(fingerprintID));
      Double[] result = manglerResults.get(manglers);

      // precision == ppv
      if ((result[Global.TRUE_POSITIVE] + result[Global.FALSE_POSITIVE]) > 0.0) {
        result[Global.PRECISION] = result[Global.TRUE_POSITIVE]
            / (result[Global.TRUE_POSITIVE] + result[Global.FALSE_POSITIVE]);
      } else {
        result[Global.PRECISION] = 0.0;
      }

      // recall == sensitivity
      if ((result[Global.TRUE_POSITIVE] + result[Global.FALSE_NEGATIVE]) > 0) {
        result[Global.RECALL] = result[Global.TRUE_POSITIVE].doubleValue()
            / (result[Global.TRUE_POSITIVE] + result[Global.FALSE_NEGATIVE]);
      } else {
        result[Global.RECALL] = 0.0;
      }

      // harmonic mean of precision and recall
      if ((result[Global.PRECISION] + result[Global.RECALL]) > 0) {
        result[Global.FSCORE] = 2 * ((result[Global.PRECISION] * result[Global.RECALL]) / (result[Global.PRECISION] + result[Global.RECALL]));
      } else {
        result[Global.FSCORE] = 0.0;
      }

      if (insert) {
        database.addTrialResults(trialID, fingerprintID_manglerID.get(fingerprintID),
            result[Global.TRUE_POSITIVE].intValue(), result[Global.TRUE_NEGATIVE].intValue(),
            result[Global.FALSE_POSITIVE].intValue(), result[Global.FALSE_NEGATIVE].intValue(),
            result[Global.PRECISION], result[Global.RECALL], result[Global.FSCORE]);
      }
      manglerResults.put(manglers, result);
    }

    finished = true;
    if (insert) {
      Global.progress.update("Trial:    Trial " + trialID + " Finished");
    } else {
      Global.progress.update("Trial:    Trial Finished");
    }
  }
}