package edu.georgetown.gucs.experiment;

/**
 * Provides the trial parameters that are needed for results comparisons; includes dataset, sample parameter, dictionary
 * parameter, min and max IDF, fingerprinter name, matcher name and minimum score, tokenizers and manglers
 * 
 * @author Lindsay Neubauer
 * 
 */
public final class TrialParameters {

  private final String dataset;
  private final String sampleParameter;
  private final String dictionaryParameter;
  private final String minIDF;
  private final String maxIDF;
  private final String tokenizers;
  private final String manglers;
  private final String fingerprinter;
  private final String matcher;
  private final String matcherScore;

  /**
   * Constructor that specifies all the parameters for a trial
   * 
   * @param dataset
   *          the string name of the dataset and the database to insert trial results
   * @param sampleParameter
   *          the string count or percent used for the file sample size in this trial; numbers less than 1 are processed
   *          as percents, numbers greater than or equal to 1 are processed as counts
   * @param dictionaryParameter
   *          the string count or percent used for dictionary creation in this trial; numbers less than 1 are processed
   *          as percents, numbers greater than or equal to 1 are processed as counts
   * @param minIDF
   *          the string minimum normalized IDF to kept in the dictionary in this trial
   * @param maxIDF
   *          the string maximum normalized IDF to kept in the dictionary in this trial
   * @param tokenizers
   *          the string containing the tokenizers used in this trial
   * @param manglers
   *          the list of different mangler strings used in this trial
   * @param fingerprinter
   *          the string name of the fingerprinter used in this trial
   * @param matcher
   *          the string name of the matcher used for fingerprint comparisons in this trial
   * @param matcherScore
   *          the string minimum score that the matcher considers a match
   */
  public TrialParameters(String dataset, String sampleParameter, String dictionaryParameter, String minIDF,
      String maxIDF, String tokenizers, String manglers, String fingerprinter, String matcher, String matcherScore) {

    this.dataset = dataset;
    this.sampleParameter = sampleParameter;
    this.dictionaryParameter = dictionaryParameter;
    this.minIDF = minIDF;
    this.maxIDF = maxIDF;
    this.tokenizers = tokenizers;
    this.manglers = manglers;
    this.fingerprinter = fingerprinter;
    this.matcher = matcher;
    this.matcherScore = matcherScore;
  }

  /**
   * Provides the dataset for this trial
   * 
   * @return the string name of the dataset and the database to insert trial results
   */
  public String getDataset() {
    return dataset;
  }

  /**
   * Provides the sample count or percent for this trial
   * 
   * @return the string count or percent used for the file sample size in this trial; numbers less than 1 are processed
   *         as percents, numbers greater than or equal to 1 are processed as counts
   */
  public String getSampleParameter() {
    if (Double.valueOf(sampleParameter) < 1) {
      return "percent=" + Double.toString((Double.valueOf(sampleParameter) * 1000));
    }
    return "count=" + sampleParameter;
  }

  /**
   * Provides the dictionary count or percent for this trial
   * 
   * @return the string count or percent used for dictionary creation in this trial; numbers less than 1 are processed
   *         as percents, numbers greater than or equal to 1 are processed as counts
   */
  public String getDictionaryParameter() {
    if (Double.valueOf(dictionaryParameter) < 1) {
      return "percent=" + Double.toString((Double.valueOf(dictionaryParameter) * 1000));
    }
    return "count=" + dictionaryParameter;
  }

  /**
   * Provides the minimum IDF for this trial
   * 
   * @return the string minimum normalized IDF to kept in the dictionary in this trial
   */
  public String getMinIDF() {
    return minIDF;
  }

  /**
   * Provides the maximum IDF for this trial
   * 
   * @return the string maximum normalized IDF to kept in the dictionary in this trial
   */
  public String getMaxIDF() {
    return maxIDF;
  }

  /**
   * Provides the fingerprinter for this trial
   * 
   * @return the string name of the fingerprinter used in this trial
   */
  public String getFingerprinter() {
    return fingerprinter;
  }

  /**
   * Provides the matcher for this trial
   * 
   * @return the string name of the matcher used for fingerprint comparisons in this trial
   */
  public String getMatcher() {
    return matcher;
  }

  /**
   * Provides the matcher minimum score for this trial
   * 
   * @return the string minimum score that the matcher considers a match
   */
  public String getMatcherScore() {
    return matcherScore;
  }

  /**
   * Provides the tokenizers for this trial
   * 
   * @return the string containing the tokenizers used in this trial
   */
  public String getTokenizers() {
    return tokenizers;
  }

  /**
   * Provides the manglers for this trial
   * 
   * @return the string containing the different manglers used in this trial
   */
  public String getManglers() {
    return manglers;
  }

  /**
   * Provides a string representation of the parameters needed for results comparisons
   * 
   * @return the string representation of the trial parameters
   */
  public String toString() {
    String parameters = dataset;
    if (Double.valueOf(sampleParameter) < 1) {
      parameters += String.format("%nsample_percent=" + (Double.valueOf(sampleParameter) * 1000));
    } else {
      parameters += String.format("%nsample_count=" + sampleParameter);
    }

    if (Double.valueOf(dictionaryParameter) < 1) {
      parameters += String.format("%ndictionary_percent=" + (Double.valueOf(dictionaryParameter) * 1000));
    } else {
      parameters += String.format("%ndictionary_count=" + dictionaryParameter);
    }

    parameters += String.format("%nIDF_range=" + minIDF + "-" + maxIDF + "%n" + tokenizers + "%n" + manglers + "%n"
        + fingerprinter + "%n" + matcher + " (" + matcherScore + ")");

    return parameters;
  }
}
