package edu.georgetown.gucs.fingerprinter;

import java.util.Iterator;
import java.util.Random;

import edu.georgetown.gucs.utility.AntlrBitSet;
import edu.georgetown.gucs.utility.ExceptionHandler;

/**
 * Creates a fingerprint that consist of a bit vector representing the presence or absence of terms in a particular
 * dictionary
 * 
 * @author Clay Shields
 **/
public class BitVectorFingerprinter extends Fingerprinter {

  /**
   * Constructor that generates the fingerprint name, version, unique identifier (GUID), system identifier, and creating
   * program. Sets the fingerprinter output to display all available information.
   */
  public BitVectorFingerprinter() {
    super();
  }

  /**
   * Constructor that loads a dictionary and its tokenizers. Generates the fingerprint name, version, unique identifier
   * (GUID), system identifier, and creating program. Sets the fingerprinter output to display all available
   * information.
   * 
   * @param dictionaryFilename
   *          the filename containing the dictionary to use for this fingerprinter
   **/
  public BitVectorFingerprinter(String dictionaryFilename) {
    super(dictionaryFilename);
  }

  /**
   * Computes a byte array fingerprint indicating the presence or absence of each token in this dictionary
   **/
  public byte[] computeFingerprint(String filename) {
    if (filename == null) {
      ExceptionHandler.handleException(new Exception("BitVectorFingerprinter: Null filename"));
    }
    if (dictionary == null) {
      ExceptionHandler.handleException(new Exception("BitVectorFingerprinter: Null dictionary"));
    }
    if (tokenizers == null) {
      ExceptionHandler.handleException(new Exception("BitVectorFingerprinter: Null tokenizer list"));
    }

    // Clear and resize the bitset
    AntlrBitSet bitvector = new AntlrBitSet(dictionary.getSize());

    // get the tokens from the tokenizer list
    Iterator<String> tokens = tokenizers.tokenizeFile(filename);
    while (tokens.hasNext()) {
      int position = dictionary.getPosition(tokens.next());
      if (position > -1) {
        bitvector.add(position);
      }
    }
    return bitvector.toBytes();
  }

  /**
   * Creates a bit vector fingerprint representing the presence or absence of terms in a particular dictionary
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename containing the dictionary to use for this fingerprinter<br>
   *          <code>args[1]</code> the filename of the document to fingerprint<br>
   *          <code>args[2]</code> the filename to write the XML output<br>
   *          <code>args[3]</code> the filename containing the output specifications
   */
  public static void main(String args[]) {
    String dictionary_file = "";
    String input_file = "";
    String output_file = "";
    String output_config = null;

    try {
      dictionary_file = args[0];
      input_file = args[1];
      if (args.length > 2)
        output_file = args[2];
      else {
        Random random = new Random();
        output_file = "fingerprint_" + random.nextInt(Integer.MAX_VALUE) + ".xml";
      }
      if (args.length > 3)
        output_config = args[3];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java BitVectorFingerprinter <dictionary file> <file to fingeprint> "
          + "<optional: destination filename (otherwise use: fingerprint_<randomInteger>.xml)> <optional: output configuration file>");
      System.exit(-1);
    }

    BitVectorFingerprinter bvfp = new BitVectorFingerprinter(dictionary_file);
    if (output_config != null)
      bvfp.outputFields(output_config);

    String fingerprint = bvfp.computeFingerprintXML(input_file);
    bvfp.generateXML(fingerprint, input_file, output_file);
    System.out.println("\n" + bvfp.getFingerprintName() + " (Base64):\n" + fingerprint);
  }
}
