package edu.georgetown.gucs.fingerprinter;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.utility.ExceptionHandler;

/**
 * Produces a {@linkplain Dictionary} in XML format from a given {@linkplain Fingerprinter} digest containing a
 * dictionary.
 * 
 * @author Lindsay Neubauer
 */
public class ExtractDictionary {

  private String dictionary_file;

  /**
   * Provides the filename of this extracted dictionary
   * 
   * @return the string filename of this extracted dictionary
   */
  public String getDictionaryFilename() {
    return dictionary_file;
  }

  /**
   * Produces an XML dictionary object from a fingerprint digest. If the <code>dictionary_file</code> is null, the
   * dictionary contained in the fingerprinter digest is used.
   * 
   * @param input_file
   *          the string name of the XML file containing the fingerprint digest
   * @param dictionary_file
   *          the string name of the XML file to write the dictionary to, if null the name of the dictionary in the
   *          fingerprint digest is used
   */
  public void extractDictionaryXML(String input_file, String dictionary_file) {
    try {
      File xml = new File(input_file);
      SAXBuilder builder = new SAXBuilder();
      Document doc = (Document) builder.build(xml);

      Element fingerprinter = doc.getRootElement();

      if (dictionary_file == null || dictionary_file.length() < 1)
        dictionary_file = fingerprinter.getChild("dictionary").getAttributeValue("name");
      else
        dictionary_file = (new File(dictionary_file)).getAbsolutePath();
      Document dictionary_xml = new Document((Element) fingerprinter.getChild("dictionary").clone());

      FileOutputStream digest = new FileOutputStream(dictionary_file);
      Format form = Format.getPrettyFormat();
      form.setEncoding("UTF-8");
      XMLOutputter out = new XMLOutputter(form);
      out.output(dictionary_xml, digest);
      System.out.println("Dictionary in " + xml.getAbsolutePath() + " written to " + dictionary_file);
      digest.close();
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleException(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleException(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleException(e);
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
    this.dictionary_file = dictionary_file;
  }

  /**
   * Extracts a dictionary from a fingerprint digest and writes it to a new dictionary file
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename containing the fingerprint and its dictionary to extract<br>
   *          <code>args[1]</code> the filename to write the extracted dictionary to
   */
  public static void main(String args[]) {
    String input_file = null;
    String output_file = null;
    try {
      input_file = args[0];
      if (args.length > 1)
        output_file = args[1];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java ExtractDictionary <input_file> <dictionary_file>");
      System.err.println("Extracts a dictionary from a fingerprint digest and writes it to a new dictionary file");
      System.exit(-1);
    }
    ExtractDictionary dictionary = new ExtractDictionary();
    dictionary.extractDictionaryXML(input_file, output_file);
  }
}