package edu.georgetown.gucs.matcher;

import edu.georgetown.gucs.fingerprinter.BitVectorFingerprinter;
import edu.georgetown.gucs.utility.AntlrBitSet;
import edu.georgetown.gucs.utility.ExceptionHandler;

/**
 * Provides a score based on a cosine similarity match to be used with a {@linkplain BitVectorFingerprinter}. Uses the
 * {@linkplain AntlrBitSet} to do the calculations.
 * 
 * @author Clay Shields
 **/
public class CosineSimilarityFingerprintMatcher extends FingerprintMatcher {

  /**
   * Constructor that sets the minimum score to use for matching two fingerprints to zero
   **/
  public CosineSimilarityFingerprintMatcher() {
    super();
  }

  /**
   * Constructor that sets the minimum score to use for matching two fingerprints
   * 
   * @param score
   *          the minimum score needed to be considered a match
   */
  public CosineSimilarityFingerprintMatcher(int score) {
    super(score);
  }

  /**
   * Constructor that sets the minimum score to use for matching two fingerprints. The string parameter is needed for
   * use with ObjectBuilder.
   * 
   * @param score
   *          the string value of the minimum score needed to be considered a match
   */
  public CosineSimilarityFingerprintMatcher(String score) {
    super(Integer.valueOf(score));
  }

  /**
   * Determines that the two fingerprints are matching if their similarity score is at or above the minimum score for
   * this fingerprinter
   */
  public boolean match(byte[] fp1, byte[] fp2) {
    if (getScore(fp1, fp2) >= minimum_score) {
      return true;
    } else {
      return false;
    }
  }

  /**
   * Determines the cosine similarity score for these two fingerprints
   */
  public int getScore(byte[] fp1, byte[] fp2) {

    if (fp1 == null || fp2 == null) {
      ExceptionHandler.handleException(new Exception("Null fingerprints given for matching"));
    }

    // Convert the byte arrays to BitSets

    AntlrBitSet matchingBits = new AntlrBitSet(fp1);
    AntlrBitSet secondBitset = new AntlrBitSet(fp2);
    int firstDegree = matchingBits.degree();
    int secondDegree = secondBitset.degree();

    // If either degree of the arrays is zero, return zero
    if (firstDegree == 0 || secondDegree == 0) {
      return 0;
    }

    // Find matching bits
    matchingBits.andInPlace(secondBitset);
    return (int) ((matchingBits.degree() / Math.sqrt(firstDegree * secondDegree)) * 99);
  }

  public static void main(String[] args) {
    String dictionary = null;
    String file1 = null;
    String file2 = null;
    try {
      dictionary = args[0];
      file1 = args[1];
      file2 = args[2];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err
          .println("Usage: java CosineSimilarityFingerprintMatcher <dictionary> <file to fingerprint> <file to fingerprint>");
      System.exit(-1);
    }

    // Create two fingerprinters
    BitVectorFingerprinter fpr = new BitVectorFingerprinter(dictionary);
    String fp1 = fpr.computeFingerprintXML(file1);
    String fp2 = fpr.computeFingerprintXML(file2);
    CosineSimilarityFingerprintMatcher fm = new CosineSimilarityFingerprintMatcher();
    System.out.println("Score: " + fm.getScoreXML(fp1, fp2));
  }
}
