package edu.georgetown.gucs.tokenizers;

import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Vector;

import net.sf.classifier4J.summariser.SimpleSummariser;
import edu.georgetown.gucs.utility.ExceptionHandler;

/**
 * Manipulates files as they are read. Settings are specified in an XML configuration file for a ParameterSearch and can
 * include:
 * <ul>
 * <li>null: makes no changes</li>
 * <li>deleteToken: deletes a given percentage of the tokens</li>
 * <li>deleteChar: deletes a given percentage of single characters</li>
 * <li>deleteWhitespace: deletes a given percentage of white space</li>
 * <li>deleteSection: deletes a given percentage of contiguous tokens</li>
 * <li>deleteHead: deletes a given percentage of contiguous tokens from the beginning of the file</li>
 * <li>deleteTail: deletes a given percentage of contiguous tokens from the end of the file</li>
 * <li>splitToken: inserts white space randomly inside a given percentage of tokens</li>
 * <li>changeToChar: changes a given percentage of characters to a random ascii lower case letter or number</li>
 * <li>changeToken: changes a given percentage of tokens to a random token from the dictionary</li>
 * <li>changeToPunc: changes a given percentage of characters to punctuation</li>
 * <li>summaryReduce: uses Classifier4J to create a document summary of a given percentage of the document size</li>
 * </ul>
 * 
 * @author Clay Shields
 **/
public class FileManglerTokenizer extends Tokenizer {

  private float deleteToken;
  private float deleteSection;
  private float deleteHead;
  private float deleteTail;
  private float deleteChar;
  private float deleteWhitespace;
  private float changeToChar;
  private float changeToPunc;
  private float changeToken;
  private float splitToken;
  private float summaryReduce;
  private Random random;
  private List<String> dictionary;
  private boolean nullProcess;
  private String tokensAsString;

  /**
   * Constructor that initializes the random number generator and clears the mangler settings
   */
  public FileManglerTokenizer() {
    random = new Random();
    nullProcess = true;
    tokensAsString = null;
    clearManglerSettings();
  }

  /**
   * Constructor that sets the mangler settings
   * 
   * @param manglers
   *          the string of mangler settings; expects non-null mangler strings in the form:
   *          "manglerName-manglerPercent", where the manglerPercent is between 1 and 99 inclusive
   */
  public FileManglerTokenizer(String manglers) {
    random = new Random();
    tokensAsString = null;
    nullProcess = false;
    parseSettings(manglers);
  }

  /**
   * Turns off (sets to zero) all the mangler settings for this tokenizer
   */
  private void clearManglerSettings() {
    deleteToken = 0;
    deleteSection = 0;
    deleteHead = 0;
    deleteTail = 0;
    deleteChar = 0;
    deleteWhitespace = 0;
    changeToChar = 0;
    changeToPunc = 0;
    changeToken = 0;
    splitToken = 0;
    summaryReduce = 0;
  }

  /**
   * Enables the mangler with the given settings in this tokenizer
   * 
   * @param manglers
   *          the string of mangler settings; expects non-null mangler strings in the form:
   *          "manglerName-manglerPercent", where the manglerPercent is between 1 and 99 inclusive
   */
  public void enableMangler(String manglers) {
    nullProcess = false;
    parseSettings(manglers);
  }

  /**
   * Enables the mangler with the given settings in this tokenizer
   * 
   * @param manglers
   *          the string of mangler settings; expects non-null mangler strings in the form:
   *          "manglerName-manglerPercent", where the manglerPercent is between 1 and 99 inclusive
   * @param tokens
   *          the list of string tokens to put through this mangler
   */
  public void enableMangler(String manglers, List<String> tokens) {
    dictionary = tokens;
    nullProcess = false;
    parseSettings(manglers);
  }

  /**
   * Disables the mangler in this tokenizer
   */
  public void disableMangler() {
    nullProcess = true;
    clearManglerSettings();
  }

  /**
   * Sets the random number generator to use with the manglers that are set in this tokenizer
   * 
   * @param random
   *          the random number generator to use with this tokenizer
   */
  public void setRNG(Random random) {
    this.random = random;
  }

  /**
   * Sets each mangler setting present in the given list; throws an error if the same mangler is present more than once
   * 
   * @param manglerSettings
   *          the string of mangler settings; expects non-null mangler strings in the form:
   *          "manglerName-manglerPercent", where the manglerPercent is between 1 and 99 inclusive
   */
  private void parseSettings(String manglers) {
    clearManglerSettings();
    if (manglers.contains("null")) {
      disableMangler();
    } else {
      String[] manglerArray = manglers.split("\\s");
      for (String manglerInput : manglerArray) {
        String[] manglerSplit = manglerInput.split("-");
        String mangler = manglerSplit[0];
        float percent = Float.valueOf(manglerSplit[1]);
        if (percent < 1 || percent > 99) {
          ExceptionHandler.handleException(new Exception("FileManglerTokenizer: percent must be between 1-99"));
        }

        if (mangler.equalsIgnoreCase("deleteToken")) {
          if (deleteToken > 0)
            ExceptionHandler.handleException(new Exception("FileManglerTokenizer: multiple deleteToken calls"));
          deleteToken = percent;
        } else if (mangler.equalsIgnoreCase("deleteSection")) {
          if (deleteSection > 0)
            ExceptionHandler.handleException(new Exception("FileManglerTokenizer: multiple deleteSection calls"));
          deleteSection = percent;
        } else if (mangler.equalsIgnoreCase("deleteHead")) {
          if (deleteHead > 0)
            ExceptionHandler.handleException(new Exception("FileManglerTokenizer: multiple deleteHead calls"));
          deleteHead = percent;
        } else if (mangler.equalsIgnoreCase("deleteTail")) {
          if (deleteTail > 0)
            ExceptionHandler.handleException(new Exception("FileManglerTokenizer: multiple deleteTail calls"));
          deleteTail = percent;
        } else if (mangler.equalsIgnoreCase("deleteChar")) {
          if (deleteChar > 0)
            ExceptionHandler.handleException(new Exception("FileManglerTokenizer: multiple deleteChar calls"));
          deleteChar = percent;
        } else if (mangler.equalsIgnoreCase("deleteWhitespace")) {
          if (deleteWhitespace > 0)
            ExceptionHandler.handleException(new Exception("FileManglerTokenizer: multiple deleteWhitespace calls"));
          deleteWhitespace = percent;
        } else if (mangler.equalsIgnoreCase("changeToChar")) {
          if (changeToChar > 0)
            ExceptionHandler.handleException(new Exception("FileManglerTokenizer: multiple changeToChar calls"));
          changeToChar = percent;
        } else if (mangler.equalsIgnoreCase("changeToken")) {
          if (changeToken > 0)
            ExceptionHandler.handleException(new Exception("FileManglerTokenizer: multiple changeToken calls"));
          changeToken = percent;
        } else if (mangler.equalsIgnoreCase("splitToken")) {
          if (splitToken > 0)
            ExceptionHandler.handleException(new Exception("FileManglerTokenizer: multiple splitToken calls"));
          splitToken = percent;
        } else if (mangler.equalsIgnoreCase("changeToPunc")) {
          if (changeToPunc > 0)
            ExceptionHandler.handleException(new Exception("FileManglerTokenizer: multiple changeToPunc calls"));
          changeToPunc = percent;
        } else if (mangler.equalsIgnoreCase("summaryReduce")) {
          if (summaryReduce > 0)
            ExceptionHandler.handleException(new Exception("FileManglerTokenizer: multiple summaryReduce calls"));
          summaryReduce = percent;
        }
      }
    }
  } // end parseSettings

  /**
   * Prints mangler settings to standard output
   */
  public void showSettings() {
    System.out.println("nullProcess: " + nullProcess);
    System.out.println("deleteToken: " + deleteToken);
    System.out.println("deleteChar: " + deleteChar);
    System.out.println("deleteWhitespace: " + deleteWhitespace);
    System.out.println("changeToChar: " + changeToChar);
    System.out.println("changeToken: " + changeToken);
    System.out.println("splitToken: " + splitToken);
    System.out.println("changeToPunc: " + changeToPunc);
    System.out.println("deleteSection: " + deleteSection);
    System.out.println("deleteHead: " + deleteHead);
    System.out.println("deleteTail: " + deleteTail);
    System.out.println("summaryReduce: " + summaryReduce);
  }

  /**
   * Alters or eliminates certain tokens based on the given mangler settings
   */
  public void tokenize(Iterator<String> iterator) {
    tokenVector.clear();
    while (iterator.hasNext()) {
      String token = iterator.next().trim();
      if (token.length() > 0)
        tokenVector.add(token);
    }

    // Apply mangler settings
    if (!nullProcess) {

      // Deletes a given percentage of the tokens
      if (deleteToken > 0) {
        int deletionSize = (int) ((deleteToken / 100) * tokenVector.size());
        for (int i = 0; i < deletionSize; i++) {
          tokenVector.remove(random.nextInt(tokenVector.size()));
        }
      }

      // Deletes a given percentage of contiguous tokens
      if (deleteSection > 0) {
        int deletionSize = (int) ((deleteSection / 100) * tokenVector.size());
        int maxStart = tokenVector.size() - deletionSize;
        if (maxStart < 0) {
          String exception = "FileManglerTokenizer: DeleteSection is larger than token vector" + ": vector size= "
              + tokenVector.size() + " setting in pct " + deleteSection + " num to delete " + deletionSize;
          ExceptionHandler.handleException(new Exception(exception));
        }
        int randomStart = (maxStart == 0) ? 0 : random.nextInt(maxStart);
        tokenVector.subList(randomStart, randomStart + deletionSize).clear();
      }

      // Deletes a given percentage of contiguous tokens from the beginning of
      // the file
      if (deleteHead > 0) {
        int deletionSize = (int) ((deleteHead / 100) * tokenVector.size());
        tokenVector.subList(0, deletionSize).clear();
      }

      // Deletes a given percentage of contiguous tokens from the end of the
      // file
      if (deleteTail > 0) {
        int deletionSize = (int) ((deleteTail / 100) * tokenVector.size());
        tokenVector.subList(tokenVector.size()-deletionSize, tokenVector.size()).clear();
      }

      // Deletes a given percentage of single characters
      if (deleteChar > 0) {
        int deletionSize = (int) ((deleteChar / 100) * tokenVector.size());
        for (int i = 0; i < deletionSize; i++) {
          StringBuffer token = new StringBuffer(tokenVector.remove(random.nextInt(tokenVector.size())));
          if (token.length() > 0) {
            token.deleteCharAt(random.nextInt(token.length()));
            tokenVector.add(token.toString());
          }
        }
      }

      // Deletes a given percentage of white space
      if (deleteWhitespace > 0) {
        int deletionSize = (int) ((deleteWhitespace / 100) * tokenVector.size());
        for (int i = 0; i < deletionSize; i++) {
          int position = random.nextInt(tokenVector.size() - 1);

          // For each gap between tokens, there is a deleteWhitespace percent
          // chance of it being deleted
          String newToken = tokenVector.get(position) + tokenVector.get(position + 1);
          tokenVector.remove(position + 1);
          tokenVector.set(position, newToken);
        }
      }

      // Changes a given percentage of characters to a random ASCII lower case
      // letter or number
      if (changeToChar > 0) {
        int deletionSize = (int) ((changeToChar / 100) * tokenVector.size());
        for (int i = 0; i < deletionSize; i++) {
          StringBuffer token = new StringBuffer(tokenVector.remove(random.nextInt(tokenVector.size())));
          token.setCharAt(random.nextInt(token.length()), (char) (random.nextInt(26) + 97));
          tokenVector.add(token.toString());
        }
      }

      // Changes a given percentage of tokens to a random token from the
      // dictionary
      if (changeToken > 0) {

        if (dictionary == null || dictionary.size() == 0) {
          ExceptionHandler.handleException(new Exception("FileManglerTokenizer: null or empty dictionary"));
        }

        int deletionSize = (int) ((changeToken / 100) * tokenVector.size());
        for (int i = 0; i < deletionSize; i++) {
          tokenVector.set(random.nextInt(tokenVector.size()), dictionary.get(random.nextInt(dictionary.size())));
        }
      }

      // Changes a given percentage of characters to punctuation
      if (changeToPunc > 0) {
        // create a vector of punctuation symbols
        Vector<Character> punc = new Vector<Character>();
        for (char i = 33; i <= 47; i++) {
          punc.add(i);
        }
        for (char i = 58; i <= 64; i++) {
          punc.add(i);
        }
        for (char i = 91; i <= 96; i++) {
          punc.add(i);
        }
        for (char i = 123; i <= 126; i++) {
          punc.add(i);
        }

        int deletionSize = (int) ((changeToPunc / 100) * tokenVector.size());
        for (int i = 0; i < deletionSize; i++) {
          StringBuffer token = new StringBuffer(tokenVector.remove(random.nextInt(tokenVector.size())));
          token.setCharAt(random.nextInt(token.length()), punc.get(random.nextInt(punc.size())));
          tokenVector.add(token.toString());
        }
      }

      // Inserts white space randomly inside a given percentage of tokens
      if (splitToken > 0) {
        int deletionSize = (int) ((splitToken / 100) * tokenVector.size());
        Vector<String> tempVector = new Vector<String>();
        for (int i = 0; i < deletionSize; i++) {
          String token = tokenVector.remove(random.nextInt(tokenVector.size()));
          if (token.length() > 1) {
            int splitPoint = random.nextInt(token.length());
            // Don't want to split the same token twice
            tempVector.add(token.substring(0, splitPoint));
            tempVector.add(token.substring(splitPoint));
          }
        }
        // Add all split tokens back into tokenVector
        tokenVector.addAll(tempVector);
      }

      // Uses Classifier4J to create a document summary of a given percentage of
      // the document size
      if (summaryReduce > 0) {
        // Turn the vector of tokens into a long string and find out how many
        // sentences there are
        int sentenceCount = tokensToSentences();

        // Figure out how many sentences to keep
        int sentenceToKeep = (int) (sentenceCount * ((100 - summaryReduce) / 100));

        SimpleSummariser summariser = new SimpleSummariser();
        String summary = summariser.summarise(tokensAsString, sentenceToKeep);

        // Split the result into a new tokenVector and replace the one we were
        // using
        Vector<String> newTokenVector = new Vector<String>();
        for (String result : summary.split("\\s")) {
          if (result.trim().length() > 0) {
            newTokenVector.add(result.trim());
          }
        }
        tokenVector = newTokenVector;
      }
    }
  }

  /**
   * Processes tokens into sentences for use with the summariser; Does special processing when first line contains
   * "Message-ID" (looks like an email); Puts period at end of each header line to clearly mark it as the end of line
   * 
   * @return the <code>int</code> number of sentences
   */
  private int tokensToSentences() {
    // This is a hack: Probably should have rewritten the File Tokenizer to do
    // this
    boolean inHeaders = (tokenVector.get(0).contains("Message-ID:")) ? true : false;

    StringBuffer fullFile = new StringBuffer(tokenVector.get(0));
    int sentenceCount = 0;

    for (int i = 1; i < tokenVector.size(); i++) {
      fullFile.append(" "+tokenVector.get(i));

      if (inHeaders) {
        // peek ahead at the next token to see if it has a :
        if ((i < tokenVector.size() - 2) && (tokenVector.get(i + 1).contains(":"))) {
          // if so, add a . to the current line
          fullFile.append(".");
          sentenceCount++;
        }
        // If at the X-FileName line should be the end of the headers
        if ((tokenVector.get(i).contains("X-FileName:"))) {
          // A token that has a . in it is probably the file name
          if (tokenVector.get(i).contains(".")) {
            inHeaders = false;
            fullFile.append(".");
            sentenceCount++;
          }
        }
      }

      // Check if we should count a new sentence
      // When we see a token that ends with a . ! or ?
      int length = tokenVector.get(i).length();

      // If it is a blank line, skip this check
      if (length != 0) {
        // char last = tokenVector.get(i).charAt(length - 1);
        if ((tokenVector.get(i).charAt(length - 1) == '.') || (tokenVector.get(i).charAt(length - 1) == '!')
            || (tokenVector.get(i).charAt(length - 1) == '?')) {
          sentenceCount++;
        }
      }
    }// end for i

    tokensAsString = fullFile.toString();
    return sentenceCount;
  }

  public static void main(String args[]) {
    FileManglerTokenizer fmt = new FileManglerTokenizer("deleteHead-20 deleteTail-20 deleteWhitespace-40");
    fmt.showSettings();

  }
}
