package edu.georgetown.gucs.tokenizers;

import java.util.Arrays;
import java.util.Dictionary;

import org.jdom.Verifier;

import edu.georgetown.gucs.utility.Pair;

/**
 * Uses Oracle's OutsideIn technology to extract text from a file. Splits contents of the file into tokens based on
 * whitespace or by line. Used for {@linkplain Dictionary} creation.
 * 
 * @author Lindsay Neubauer
 */
public class OutsideInFileTokenizer extends FileTokenizer {

  /**
   * Constructor that sets the token creation mode to split based on whitespace
   */
  public OutsideInFileTokenizer() {
    mode = "tokens";
  }

  /**
   * Constructor that sets the token creation mode. If set to "lines" the file is split by line; if set to "tokens" the
   * file is split based on whitespace.
   * 
   * @param mode
   *          the string mode to split tokens based on whitespace or by line
   */
  public OutsideInFileTokenizer(String mode) {
    this.mode = mode;
  }

  public void tokenize(String filename) {
    readFile(filename);
  }

  /**
   * Uses a JNI to call Oracle's Outside In API to extract text from a file and splits that text into tokens
   * 
   * @param filename
   *          the string filename of the document to split into tokens
   */
  protected synchronized void readFile(String filename) {
    String tokens[] = new String[] {};
    tokenVector.clear();

    if (mode.equals("lines")) {
      tokens = extractText(filename, 1);
    } else {
      tokens = extractText(filename, 0);
    }

    if (tokens != null && tokens.length > 0) {
      int positionVector[] = getPositions();
      int first = positionVector[0];
      int second;
      for (int i = 1; i < positionVector.length; i++) {
        second = positionVector[i];
        if (first <= second) {
          positions.add(new Pair<Integer, Integer>(first, second));
          first = second;
        }
      }
      tokenVector.addAll(Arrays.asList(tokens));
      for (int i = 0; i < tokenVector.size(); i++) {
        // Remove whitespace-only entries
        if (tokenVector.get(i).trim().length() == 0) {
          tokenVector.remove(i);
          positions.remove(i);
          i--;
        } else {
          StringBuffer tokenName = new StringBuffer();
          for (int j = 0; j < tokenVector.get(i).length(); j++) {
            if (Verifier.isXMLCharacter(tokenVector.get(i).charAt(j)))
              tokenName.append(tokenVector.get(i).charAt(j));
          }
          // Remove leading and trailing whitespace
          tokenVector.set(i, tokenName.toString().trim());
        }
      }
    } else {
      System.err.println("tokenizeOutsideIn: tokens from JNI is null");
      // ExceptionHandler.handleException(new
      // Exception("tokenizeOutsideIn: tokens from JNI is null"));
    }

    // These are the tokens (by line) from before the first page of the file
    /*
     * Vector<String> metaDataVector = new Vector<String>(Arrays.asList(getMetaData())); for (int m = 0; m <
     * metaDataVector.size() && metaDataVector.get(m) != null; m++) { // Remove whitespace-only entries if
     * (metaDataVector.get(m).trim().length() == 0) { metaDataVector.remove(m); m--; } else { // Remove leading and
     * trailing whitespace metaDataVector.set(m, metaDataVector.get(m).trim());
     * System.out.println("metaDataVector["+m+"] = "+metaDataVector.get(m)); } }
     */
  }

  /**
   * Prints to the screen a portion of the given document
   */
  public void checkIndexing(String filename, int start_index, int end_index) {
    getIndex(filename, start_index, end_index);
  }

  /**
   * Prints to the screen a portion of the given document
   * 
   * @param filename
   *          the string filename of the document to split into tokens
   * @param start
   *          the number of characters to skip in file before printing, inclusive
   * @param end
   *          the number of characters to print from the file, exclusive
   */
  private native void getIndex(String filename, int start, int end);

  /**
   * Extracts text from a file and splits it into tokens based on whitespace or by line
   * 
   * @param filename
   *          the string filename of the document to split into tokens
   * @param lines
   *          the <code>int</code> that specifies whether to split tokens based on whitespace (lines=0) or by line
   *          (lines=1)
   * 
   * @return the string array containing the tokens from the file
   */
  private native synchronized String[] extractText(String filename, int lines);

  /**
   * Provides an integer array of the position of each token. The returned array is initialized and populated in the
   * <code>extractText</code> method and must only be called after a successful call to that method.
   * 
   * @return the <code>int</code> array of positions of each token
   */
  private native int[] getPositions();

  /**
   * Returns meta data tokens from the text extracted. The returned array is initialized and populated in the
   * <code>extractText</code> method and must only be called after a successful call to that method.
   * 
   * @return the string array containing the meta data tokens from the file
   */
  // private native synchronized String[] getMetaData();

  /**
   * Tokenizes a file using OutsideIn and prints the resulting tokens to the the screen
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename of the document to split into tokens
   */
  public static void main(String[] args) {
    try {
      String filename = args[0];
      OutsideInFileTokenizer outIn = new OutsideInFileTokenizer();
      outIn.tokenize(filename);
      printTokens(outIn.iterator());
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java OutsideInFileTokenizer <file name>");
      System.err.println("Tokenizes a file using OutsideIn and prints the resulting tokens to the the screen");
      System.exit(-1);
    }
  }

  static {
    System.loadLibrary("OutsideInFileTokenizer");
  }
}
