package edu.georgetown.gucs.tokenizers;

import java.util.Iterator;

import edu.georgetown.gucs.utility.Pair;

/**
 * Eliminates tokens nested inside markup language tags; assumes that tokens have been split by line rather than using
 * whitespace
 * 
 * @author Lindsay Neubauer
 * 
 */
public class StripMarkupTokenizer extends Tokenizer {

  private boolean keepScript = false;

  public StripMarkupTokenizer() {
  }

  /**
   * Constructor that specifies whether to keep tokens nested inside <code><\script></code> tags; the default is to
   * eliminate these tokens
   * 
   * @param keepScript
   *          the string value (true or false) specifying whether to keep tokens nested inside <code><\script></code>
   *          tags
   */
  public StripMarkupTokenizer(String keepScript) {
    this.keepScript = Boolean.valueOf(keepScript);
  }

  /**
   * Eliminates tokens nested inside markup language tags; assumes that tokens have been split by line rather than using
   * whitespace
   */
  public void tokenize(Iterator<String> iterator) {
    tokenVector.clear();
    boolean script = false;
    while (iterator.hasNext()) {
      for (String token : iterator.next().split(">")) {

        // TODO: Printing error message instead of throwing exception
        // because govdocs has unbalanced brackets
        if (token.indexOf("<") != token.lastIndexOf("<")) {
          System.err.println("Unbalanced brackets error: " + token + ">");
        }
        if (token.contains("script") && token.contains("<")) {
          script = !script;
        }

        if (!script) {
          if (token.indexOf('<') > 0) {
            tokenVector.add(token.substring(0, token.indexOf('<')));
          } else if (token.indexOf('<') < 0) {
            tokenVector.add(token);
          }
        } else if (script && keepScript) {
          tokenVector.add(token);
        }
      }
    }
  }

  /**
   * Eliminates tokens nested inside markup language tags; assumes that tokens have been split by line rather than using
   * whitespace
   */
  public void tokenize(Iterator<String> tokensIterator, Iterator<Pair<Integer, Integer>> positionsIterator) {
    tokenVector.clear();
    positions.clear();
    boolean script = false;
    while (tokensIterator.hasNext() && positionsIterator.hasNext()) {
      Pair<Integer, Integer> pos = positionsIterator.next();
      for (String token : tokensIterator.next().trim().split(">")) {

        // TODO: Printing error message instead of throwing exception
        // because govdocs has unbalanced brackets
        if (token.indexOf("<") != token.lastIndexOf("<")) {
          System.err.println("Unbalanced brackets error: " + token + ">");
        }
        if (token.contains("script") && token.contains("<")) {
          script = !script;
        }

        if (!script) {
          if (token.indexOf('<') > 0) {
            tokenVector.add(token.substring(0, token.indexOf('<')));
            positions.add(pos);
          } else if (token.indexOf('<') < 0) {
            tokenVector.add(token);
            positions.add(pos);
          }
        } else if (script && keepScript) {
          tokenVector.add(token);
          positions.add(pos);
        }
      }
    }
  }
}
