package edu.georgetown.gucs.tokenizers;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import edu.georgetown.gucs.utility.ExceptionHandler;

/**
 * Tests tokenizers using a configuration file that contains a list of {@linkplain Tokenizer} objects to use on a
 * document and prints the results of the tokenization
 * 
 * @author Clay Shields
 * 
 */
public class TokenizeFile extends Tokenizer {

  private Vector<String> tokenizers;

  /**
   * Constructor that initializes an empty string vector of tokenizers to use
   */
  public TokenizeFile() {
    tokenizers = new Vector<String>();
  }

  /**
   * Constructor that sets the tokenizers from a configuration file to use on this document. The first tokenizer in the
   * configuration file must support reading from a file.
   * 
   * @param config_file
   *          the string XML filename that specifies these tokenizers
   */
  public TokenizeFile(String config_file) {
    tokenizers = new Vector<String>();
    addTokenizers(config_file);
  }

  /**
   * Sets the tokenizers to use on this document
   * 
   * @param tokenVec
   *          the string vector containing these tokenizers
   */
  public void setTokenizers(Vector<String> tokenVec) {
    tokenizers = tokenVec;
  }

  /**
   * Provides the list of tokenizers to use on this document
   * 
   * @return the string vector containing these tokenizers
   */
  public Vector<String> getTokenizers() {
    return tokenizers;
  }

  /**
   * Adds tokenizers from an XML file to the list of tokenizers to use on this document
   * 
   * @param config_file
   *          the string XML filename that specifies these tokenizers
   */
  @SuppressWarnings("unchecked")
  public void addTokenizers(String config_file) {
    try {
      File xml = new File(config_file);
      SAXBuilder builder = new SAXBuilder();
      Document doc = (Document) builder.build(xml);
      Element dict = doc.getRootElement();

      for (Element tokenizerElement : (List<Element>) dict.getChildren("tokenizer")) {
        String tokenizer = tokenizerElement.getAttributeValue("name");
        if (tokenizers.size() == 0 && !tokenizer.contains("FileTokenizer")) {
          ExceptionHandler.handleException(new Exception("Dictionary: First tokenizer (" + tokenizer
              + ") must read from a file."));
        }

        if (tokenizerElement.getAttributeValue("parameter") != null)
          tokenizer += " " + (tokenizerElement.getAttributeValue("parameter"));
        tokenizers.add(tokenizer.trim());
      }
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleException(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleException(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleException(e);
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Splits the given file into tokens and alters or eliminates those tokens based on the vector of tokenizers. Prints
   * each token to standard output. The first tokenizer in the vector of tokenizers must support reading from a file.
   * 
   * @param filename
   *          the name of the file to be tokenized
   * @param tokenizerNames
   *          the string vector containing the tokenizers to use
   */
  public void tokenize(String filename, Vector<String> tokenizerNames) {
    TokenizerList tokenizerList = new TokenizerList(tokenizerNames);
    Iterator<String> tokens = tokenizerList.tokenizeFile(filename);
    printTokens(tokens);
  }

  /**
   * Tokenizes a text file and prints the resulting tokens to the the screen
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename of the document to split into tokens<br>
   *          <code>args[1]</code> the filename of the tokenizer configuration file
   */
  public static void main(String args[]) {
    String input_file = "";
    String config_file = "";
    try {
      input_file = args[0];
      config_file = args[1];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java -jar TokenizeFile.jar <file name> <tokenizers config file>");
      System.err.println("Tokenizes a text file and prints the resulting tokens to the the screen");
      System.exit(-1);
    }
    TokenizeFile newTokes = new TokenizeFile(config_file);
    newTokes.tokenize(input_file, newTokes.getTokenizers());
  }
}
