package edu.georgetown.gucs.tokenizers;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Pair;

/**
 * Splits a document into tokens (either by line or by word) or alters the tokens in various ways. This is the base
 * class for all tokenizers.
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 */
public class Tokenizer {

  /** The list of each token in order of its appearance */
  protected Vector<String> tokenVector = new Vector<String>();

  /** The list of integer pairs corresponding to the position of each token */
  protected Vector<Pair<Integer, Integer>> positions = new Vector<Pair<Integer, Integer>>();

  /**
   * Provides the list of each token in order of its appearance
   * 
   * @return the list containing each token as a string
   */
  public List<String> getTokenVector() {
    return tokenVector;
  }

  /**
   * Provides the list of integer pairs corresponding to the position of each token
   * 
   * @return the list of integer pairs
   */
  public List<Pair<Integer, Integer>> getPositionsVector() {
    return positions;
  }

  /**
   * Alters or eliminates certain tokens. By default, this method does nothing and must be over-ridden in each specific
   * tokenizer.
   * 
   * @param iterator
   *          the string iterator over the token elements
   */
  public void tokenize(Iterator<String> iterator) {
  }

  /**
   * Alters or eliminates certain tokens when using Splitter. By default, this method does nothing and must be
   * over-ridden in each specific tokenizer.
   * 
   * @param tokensIterator
   *          the string iterator over the token elements
   * @param positionsIterator
   *          the integer pair iterator over the start and end position elements
   */
  public void tokenize(Iterator<String> tokensIterator, Iterator<Pair<Integer, Integer>> positionsIterator) {
  }

  /**
   * Splits the document into tokens. By default, this method calls the ExceptionHandler to throw a
   * "Tokenizer does not support reading from a file" exception. It must be over-ridden in each tokenizer that supports
   * reading from a file.
   * 
   * @param filename
   *          the filename of the document to split into tokens
   */
  public void tokenize(String filename) {
    ExceptionHandler.handleException(new Exception("Tokenizer does not support reading from a file"));
  }

  /**
   * Provides an iterator over the tokens in this tokenizer
   * 
   * @return the iterator over the string tokens
   */
  public Iterator<String> iterator() {
    return tokenVector.iterator();
  }

  /**
   * Provides an iterator over the positions of elements in this tokenizer
   * 
   * @return the iterator over the integer pairs of start and end position elements
   */
  public Iterator<Pair<Integer, Integer>> position_iterator() {
    return positions.iterator();
  }

  /**
   * Prints each token to the system output stream using the UTF-8 charset
   * 
   * @param tokens
   *          the string iterator over the tokens in this tokenizer
   */
  public static void printTokens(Iterator<String> tokens) {
    try {
      // Not sure this "UTF-8" does anything
      OutputStreamWriter out = new OutputStreamWriter(System.out, "UTF-8");
      BufferedWriter writer = new BufferedWriter(out);
      while (tokens.hasNext()) {
        writer.write(tokens.next() + "\r\n");
      }
      out.flush();
      writer.flush();
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
  }
}
