package edu.georgetown.gucs.tokenizers;

import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Vector;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.ObjectBuilder;
import edu.georgetown.gucs.utility.Pair;

/**
 * An ordered list of {@linkplain Tokenizer} objects to split a document into tokens and alter the tokens in various
 * ways. The first Tokenizer in this list needs to be able to read directly from a file.
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 */
public class TokenizerList extends Tokenizer implements Serializable {

  private static final long serialVersionUID = -1191618315036978305L;
  private List<String> names;

  // First tokenizer in this list needs to be able to read directly from a file
  private List<Tokenizer> tokenizers;

  /**
   * Constructor that initializes an empty list of tokenizers
   */
  public TokenizerList() {
    tokenizers = new Vector<Tokenizer>();
    names = new Vector<String>();
  }

  /**
   * Constructor that takes a list of the tokenizer names
   * 
   * @param tokenizerNames
   *          the string names of the tokenizers to use; the first tokenizer must be able to read from a file
   */
  public TokenizerList(List<String> tokenizerNames) {
    tokenizers = new Vector<Tokenizer>();
    names = new Vector<String>();
    addTokenizerNames(tokenizerNames);
  }

  /**
   * Writes the non-static and non-transient fields of this class and the string tokenizer names to the output stream
   * 
   * @param out
   *          the objectOutputStream to write this tokenizerList object to
   */
  private void writeObject(ObjectOutputStream out) {
    try {
      out.defaultWriteObject();

      for (String name : names) {
        out.writeObject(name);
      }
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Reads the non-static and non-transient fields of this class and the string tokenizer names to the output stream
   * 
   * @param in
   *          the objectInputStream to read this tokenizerList object from
   */
  private void readObject(ObjectInputStream in) {
    try {
      in.defaultReadObject();
      String name;
      while ((name = (String) in.readObject()) != null) {
        names.add(name);
      }
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    } catch (ClassNotFoundException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Enables the mangler with the given settings in the FileManglerTokenizer object in this list
   * 
   * @param settings
   *          the string containing the mangler settings
   */
  public void enableMangler(String settings) {
    FileManglerTokenizer fmt = (FileManglerTokenizer) tokenizers.get(findFileMangler());
    fmt.enableMangler(settings);
  }

  /**
   * Enables the mangler with the given settings in the FileManglerTokenizer object in this list
   * 
   * @param settings
   *          the string containing the mangler settings
   * @param tokens
   *          the string list containing the tokens to put through this mangler
   */
  public void enableMangler(String settings, List<String> tokens) {

    if (tokens == null) {
      System.err.println("TokenizerList: Null list of tokens to mangle");
    }

    FileManglerTokenizer fmt = (FileManglerTokenizer) tokenizers.get(findFileMangler());
    fmt.enableMangler(settings, tokens);
  }

  /**
   * Disables the mangler in the FileManglerTokenizer object in this list
   */
  public void disableMangler() {
    FileManglerTokenizer fmt = (FileManglerTokenizer) tokenizers.get(findFileMangler());
    fmt.disableMangler();
  }

  /**
   * Sets the random number generator to use with a FileManglerTokenizer object in this list
   * 
   * @param random
   *          the random number generator to use if there is a fileManglerTokenizer object in this list
   */
  public void setManglerRNG(Random random) {
    FileManglerTokenizer fmt = (FileManglerTokenizer) tokenizers.get(findFileMangler());
    fmt.setRNG(random);
  }

  /**
   * Finds index of FileManglerTokenizer in this list of tokenizers. Assumes there is a FileManglerTokenizer in this
   * list; otherwise an error occurs.
   * 
   * @return the index of the FileManglerTokenizer
   */
  private int findFileMangler() {
    int index = names.indexOf("FileManglerTokenizer");
    if (index == -1)
      ExceptionHandler.handleException(new Exception("TokenizerList: FileManglerTokenizer not found"));
    return index;
  }

  /**
   * Adds a tokenizer to the end of this list of tokenizers. Uses reflection to create the correct object. Allows for
   * runtime specification of tokenizer type.
   * 
   * @param tokenizerName
   *          the string name of the tokenizer to be added to the end of this list
   */
  public void addTokenizer(String tokenizerName) {
    Tokenizer toInsert = (Tokenizer) ObjectBuilder.buildObject(tokenizerName);
    tokenizers.add(toInsert);
    names.add(tokenizerName);
  }

  /**
   * Adds tokenizers to the end of this list of tokenizers
   * 
   * @param tokenizerNames
   *          the list of tokenizers names to use; the first tokenizer must be able to read from a file
   */
  public void addTokenizerNames(List<String> tokenizerNames) {
    Iterator<String> nameIterator = tokenizerNames.iterator();
    while (nameIterator.hasNext()) {
      String nextName = nameIterator.next();
      addTokenizer(nextName);
    }
  }

  /**
   * Clears this list of tokenizers
   **/
  public void clear() {
    tokenizers.clear();
    names.clear();
  }

  /**
   * Provides the ordered list of the tokenizer names in this list
   * 
   * @return the list of string tokenizer names
   */
  public List<String> getNames() {
    return names;
  }

  /**
   * Provides the ordered list of the tokenizers in this list
   * 
   * @return the list of tokenizers
   */
  public List<Tokenizer> getTokenizers() {
    return tokenizers;
  }

  /**
   * Prints the ordered list of the tokenizer names in this list
   */
  public void printTokenizerNames() {
    Iterator<String> nameIterator = names.iterator();
    while (nameIterator.hasNext()) {
      System.out.println("     " + nameIterator.next());
    }
  }

  /**
   * Applies each tokenizer from this list, in order, on the file; the first tokenizer must be able to read from a file
   * and the tokenizers must already be instantiated
   * 
   * @param filename
   *          the string name of the file to tokenize
   * 
   * @return the string iterator over the tokens remaining after using each tokenizer in this list
   */
  public Iterator<String> tokenizeFile(String filename) {
    Iterator<Tokenizer> tokenizerIterator = tokenizers.iterator();
    if (tokenizerIterator.hasNext()) {
      Tokenizer first = tokenizerIterator.next();
      first.tokenize(filename);
      return tokenize();
    } else {
      ExceptionHandler.handleException(new Exception("Null list of tokenizers in TokenizerList"));
    }
    return null;
  }

  /**
   * Applies each tokenizer from this list, in order, on the file; the first tokenizer must be able to read from a file
   * and the tokenizers must already be instantiated
   * 
   * @param file
   *          the file to tokenize
   * 
   * @return the string iterator over the tokens remaining after using each tokenizer in this list
   */
  public Iterator<String> tokenizeFile(File file) {
    return tokenizeFile(file.getAbsolutePath());
  }

  /**
   * Applies each tokenizer from this list, in order, on the tokens; the first tokenizer must be able to read from a
   * file and create the list of tokens
   * 
   * @return the string iterator over the tokens remaining after using each tokenizer in this list
   */
  public Iterator<String> tokenize() {
    // iterator over the List of Tokenizer objects
    Iterator<Tokenizer> tokenizerIterator = tokenizers.iterator();

    // if there's at least one tokenizer
    if (tokenizerIterator.hasNext()) {
      // get the first tokenizer from this list
      Tokenizer first = tokenizerIterator.next();

      // get an iterator from the first tokenizer
      Iterator<String> tokens = first.iterator();

      if (tokens.hasNext()) {
        // if there are more tokenizers, feed the current tokenVector iterator
        // into the next tokenizer
        Tokenizer current;
        while (tokenizerIterator.hasNext()) {
          current = tokenizerIterator.next();
          current.tokenize(tokens);
          tokens = current.iterator();
        }
        // return the remaining tokens after going through each tokenizer in
        // this list
        return tokens;
      } else {
        // TODO: Throwing an exception here messes up OutsideInFileTokenizer -
        // but may be need for
        // something else. Check it out to makes sure.
        // Do we want to throw an exception? Its possible that our tokenizers
        // eliminate all tokens from a file

        // ExceptionHandler.handleException(new
        // Exception("TokenizerList: Trying to tokenize an empty tokenizer."));
      }
    }
    // else, we have a null list, handle exception: it seems that we should only
    // get here if we did
    // not call tokenizeFile(String), e.g., calling tokenize() first
    else {
      ExceptionHandler.handleException(new Exception("Null list of tokenizers in TokenizerList"));
    }
    return null;
  }

  /**
   * Applies each tokenizer from this list, in order, on the tokens provided by a Splitter object; no tokenizers that
   * are able to read from a file should be present in this list
   * 
   * @param tokens_iterator
   *          the string iterator over the token elements
   * @param positions_iterator
   *          the integer pair iterator over the start and end position elements
   * 
   * @return the pair iterator over the tokens and their position pair that remain after using each tokenizer in this
   *         list
   */
  protected Pair<Iterator<String>, Iterator<Pair<Integer, Integer>>> split_tokenize(Iterator<String> tokens_iterator,
      Iterator<Pair<Integer, Integer>> positions_iterator) {
    // iterator over the List of Tokenizer objects
    Iterator<Tokenizer> tokenizerIterator = tokenizers.iterator();

    // if there are tokenizers, feed the current tokenVector iterator into the
    // next tokenizer
    Tokenizer current;
    while (tokenizerIterator.hasNext()) {
      current = tokenizerIterator.next();
      current.tokenize(tokens_iterator, positions_iterator);
      tokens_iterator = current.iterator();
      positions_iterator = current.position_iterator();
    }
    // return the remaining tokens and positions after going through each
    // tokenizer in this list
    return new Pair<Iterator<String>, Iterator<Pair<Integer, Integer>>>(tokens_iterator, positions_iterator);
  }

  /**
   * Provides the string iterator over this document's tokens from the last tokenizer from this list
   * 
   * @return the iterator over the string tokens from this document
   */
  public Iterator<String> getIterator() {
    if (tokenizers == null) {
      ExceptionHandler.handleException(new Exception("TokenizerList: Null list of tokenizers"));
    }

    // Find the last tokenizer in the list and return the iterator from that
    Tokenizer tokens = null;
    Iterator<Tokenizer> tokenizerIterator = tokenizers.iterator();
    while (tokenizerIterator.hasNext()) {
      tokens = tokenizerIterator.next();
    }
    return tokens.iterator();
  }

  /*
   * public static void main ( String args[] ) {
   * 
   * String filename = args[0];
   * 
   * Vector<String> names = new Vector<String>(); names.add("FileTokenizer tokens"); names.add("FileManglerTokenizer");
   * names.add("StripPunctuationTokenizer"); names.add("PorterTokenizer"); names.add("MinimumLengthTokenizer(3)");
   * names.add("MaximumLengthTokenizer(10)"); names.add("RemoveNumericTokensTokenizer");
   * 
   * TokenizerList tokenizers = new TokenizerList(names);
   * 
   * Vector<String> out = tokenizers.getNames(); System.out.println("Tokenizers: ==================>"); for (String
   * output : out) { System.out.println(output); }
   * 
   * Vector<String> manglerSetting = new Vector<String>(); manglerSetting.add("deletehead 20");
   * manglerSetting.add("changetochar 20"); tokenizers.enableMangler(manglerSetting);
   * System.out.println("\nMangled Tokens: ==================>"); printTokens(tokenizers.tokenizeFile(filename));
   * 
   * tokenizers.disableMangler(); System.out.println("\nUnmangled Tokens: ==================>");
   * printTokens(tokenizers.tokenizeFile(filename)); }
   */
}
