package sdtext;

import java.util.Random;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.dictionary.ShowDictionaryStatistics;
import edu.georgetown.gucs.dictionary.ShowDictionaryTokens;
import edu.georgetown.gucs.dictionary.TrimDictionary;
import edu.georgetown.gucs.experiment.Experiment;
import edu.georgetown.gucs.fingerprinter.BitVectorFingerprinter;
import edu.georgetown.gucs.fingerprinter.ExtractDictionary;
import edu.georgetown.gucs.matcher.CompareDirectory;
import edu.georgetown.gucs.matcher.ScoreFingerprints;
import edu.georgetown.gucs.tokenizers.ArabicFileTokenizer;
import edu.georgetown.gucs.tokenizers.ChineseFileTokenizer;
import edu.georgetown.gucs.tokenizers.OutsideInFileTokenizer;
import edu.georgetown.gucs.tokenizers.TokenizeFile;
import edu.georgetown.gucs.utility.ExceptionHandler;

/**
 * Provides a command line interface to run the different programs within SDTEXT (Similarity Digest Text)
 * 
 * @author Lindsay Neubauer
 * 
 */
public class CommandLine {

  private enum Program {
    HELP("SDTEXT", "Usage: java -jar sdtext.jar PROGRAM [OPTION]...\r\n" + "Run PROGRAM with specified OPTION(s).\r\n"
        + "Example: sdtext TokenizeFile -i document.txt -c tokenizers.xml\r\n"
        + "Example: sdtext TokenizeFile --input document.txt --config tokenizers.xml", "\r\nOPTIONS:\r\n\r\n"
        + "-i, --input          filename to read for input\r\n"
        + "-c, --config         filename of configuration file\r\n"
        + "-o, --output         filename to write XML output\r\n"
        + "-d, --dictionary     filename containing dictionary\r\n"
        + "-f, --fingerprint    filename containing fingerprint\r\n" + "-p, --directory      path to a directory\r\n"
        + "-m, --matcher        name of matcher to use\r\n" + "-n, --dataset        name of database to use\r\n"
        + "-b, --min-idf        minimum normalized IDF (0.0 - 1.0) to consider in dictionary, default = 0.0\r\n"
        + "-t, --max-idf        maximum normalized IDF (0.0-1.0) to consider in dictionary, default = 1.0\r\n"
        + "-s, --min-score      minimum score (0-99) needed to output, default = 0\r\n"
        + "-a, --action         \"show\" matches on screen or \"insert\" matches into database\r\n"
        + "-h, --help           display help menu"),

    TOKENIZEFILE("TokenizeFile", "Tokenizes a text file and prints the resulting tokens to the the screen",
        "-i, --input          filename of document to split into tokens\r\n"
            + "-c, --config         filename of tokenizer configuration file"),

    ARABICFILETOKENIZER("ArabicFileTokenizer",
        "Tokenizes an Arabic text file and prints the resulting tokens to the the screen",
        "-i, --input          filename of Arabic language document to split into tokens"),

    CHINESEFILETOKENIZER("ChineseFileTokenizer", "Tokenizes a Chinese text file into single characters and using "
        + "a probabilistic model and prints both sets of resulting tokens to the the screen",
        "-i, --input          filename of Chinese language document to split into tokens"),
        
    OUTSIDEINFILETOKENIZER("OutsideInFileTokenizer", "Tokenizes a file using OutsideIn and prints the resulting tokens to the the screen",
        "-i, --input          filename of document to split into tokens"),

    CREATEDICTIONARY("CreateDictionary",
        "Creates a dictionary from the specified directory using the provided tokenizers",
        "-c, --config         filename of the tokenizer configuration file\r\n"
            + "-p, --directory      directory to create the dictionary from\r\n"
            + "-o, --output         filename to write the dictionary"),

    SHOWDICTIONARYSTATISTICS("ShowDictionaryStatistics", "Prints statistics for a dictionary",
        "-d, --dictionary     filename containing the dictionary to show"),

    SHOWDICTIONARYTOKENS("ShowDictionaryTokens",
        "Prints dictionary tokens and their frequencies, IDFs, and normalized IDFs",
        "-d, --dictionary     filename containing the dictionary to show"),

    TRIMDICTIONARY(
        "TrimDictionary",
        "Trims a dictionary by removing any token that is outside a range of normalized IDFs",
        "-d, --dictionary     filename containing the dictionary to trim\r\n"
            + "-b, --min-idf        minimum normalized IDF (0.0 - 1.0) to keep in trimmed dictionary, default = 0.0\r\n"
            + "-t, --max-idf        maximum normalized IDF (0.0 - 1.0) to keep in trimmed dictionary, default = 1.0\r\n"
            + "-o, --output         optional: filename to write the trimmed dictionary"),

    BITVECTORFINGERPRINTER("BitVectorFingerprinter",
        "Creates a bit vector fingerprint representing the presence or absence of terms in a particular dictionary",
        "-i, --input          filename of the document to fingerprint\r\n"
            + "-d, --dictionary     filename containing the dictionary to use\r\n"
            + "-c, --config         optional: filename containing the output configuration file\r\n"
            + "-o, --output         optional: filename to write the XML output"),

    SCOREFINGERPRINTS("ScoreFingerprints",
        "Compares two XML fingerprints and provides a score for their degree of similarity",
        "-m, --matcher        matcher name to score the fingerprints\r\n"
            + "-f, --fingerprint    filename containing fingerprint\r\n"
            + "-f, --fingerprint    filename containing fingerprint"),

    EXTRACTDICTIONARY("ExtractDictionary",
        "Extracts a dictionary from a fingerprint digest and writes it to a new dictionary file",
        "-i, --input          filename containing the fingerprint and its dictionary to extract\r\n"
            + "-o, --output         filename to write the extracted dictionary"),

    COMPAREDIRECTORY("CompareDirectory",
        "Outputs a list of files and their scores, comparing a given fingerprint to a directory of files",
        "-m, --matcher        matcher name to make comparison\r\n"
            + "-f, --fingerprint    filename containing fingerprint\r\n"
            + "-p, --directory      directory path to compare\r\n"
            + "-s, --min-score      optional: minimum score (0-99) needed to output a file, default = 0\r\n"
            + "-d, --dictionary     optional: filename containing the dictionary to use"),

    EXPERIMENT(
        "Experiment",
        "Outputs results from multiple runs with different experiment settings and optionally stores them in a database",
        "-c, --config         filename of the experiment configuration file");

    private final String name;
    private final String use;
    private final String options;

    /**
     * Stores information about each program
     * 
     * @param name
     *          the string name of the program
     * @param use
     *          the string explaining the purpose of the program
     * @param options
     *          the string specifying the command line options for the program
     */
    Program(String name, String use, String options) {
      this.name = name;
      this.use = use;
      this.options = options;
    }
  }

  private static String input = null; // -i, --input
  private static String output = null; // -o, --output
  private static String config = null; // -c, --config
  private static String dictionary = null; // -d, --dictionary
  private static String directory = null; // -p, --directory
  private static String fingerprint = null; // -f, --fingerprint
  private static String fingerprint2 = null;
  private static String matcher = null; // -m, --matcher
  private static double min_idf = 0.0; // -b, --min-idf
  private static double max_idf = 1.0; // -t, --max-idf
  private static int min_score = 0; // -s, --min-score

  /**
   * Prints SDTEXT help screen, including all programs and their options
   */
  private static void printHelp() {
    boolean print_programs = false;
    for (Program program : Program.values()) {
      if (print_programs) {
        System.err.println();
        System.err.println("PROGRAMS:");
        System.err.println();
        print_programs = false;
      }
      System.err.println(program.name + ": " + program.use);
      System.err.println(program.options);
      System.err.println();
      if (program == Program.HELP)
        print_programs = true;
    }
  }

  /**
   * Prints program name, use, and command line options, to be used when running a program that is missing a required
   * command line option
   * 
   * @param program
   *          the program to print error message for
   */
  private static void printError(Program program) {
    System.err.println("SDTEXT: Missing required OPTION");
    System.err.println(program.name + ": " + program.use);
    System.err.println(program.options);
  }

  /**
   * Runs the specified SDTEXT program
   * 
   * @param program
   *          the program to execute
   */
  private static void sdtext(Program program) {

    switch (program) {
    case TOKENIZEFILE:
      if (config != null && input != null) {
        TokenizeFile tokenizeFile = new TokenizeFile(config);
        tokenizeFile.tokenize(input, tokenizeFile.getTokenizers());
      } else {
        printError(program);
      }
      break;
    case ARABICFILETOKENIZER:
      if (input != null) {
        ArabicFileTokenizer arabicTokenizer = new ArabicFileTokenizer();
        arabicTokenizer.tokenize(input);
        ArabicFileTokenizer.printTokens(arabicTokenizer.iterator());
      } else {
        printError(program);
      }
      break;
    case CHINESEFILETOKENIZER:
      if (input != null) {
        ChineseFileTokenizer chineseTokenizer = new ChineseFileTokenizer();
        chineseTokenizer.setMode("individual");
        System.out.println("---------------------------");
        System.out.println("Mode: Individual Characters");
        System.out.println("---------------------------");
        chineseTokenizer.tokenize(input);
        ChineseFileTokenizer.printTokens(chineseTokenizer.iterator());
        chineseTokenizer.setMode("smart");
        System.out.println("------------------------------------------------");
        System.out.println("Mode: Smart/Probabalistic Character Combinations");
        System.out.println("------------------------------------------------");
        chineseTokenizer.tokenize(input);
        ChineseFileTokenizer.printTokens(chineseTokenizer.iterator());
      } else {
        printError(program);
      }
      break;
    case OUTSIDEINFILETOKENIZER:
      if (input != null) {
        OutsideInFileTokenizer outIn = new OutsideInFileTokenizer();
        outIn.tokenize(input);
        OutsideInFileTokenizer.printTokens(outIn.iterator());
      } else {
        printError(program);
      }
      break;
    case CREATEDICTIONARY:
      if (config != null && output != null && directory != null) {
        Dictionary dict = new Dictionary();
        dict.makeDictionary(config, directory, output);
      } else {
        printError(program);
      }
      break;
    case SHOWDICTIONARYSTATISTICS:
      if (dictionary != null) {
        new ShowDictionaryStatistics(dictionary);
      } else {
        printError(program);
      }
      break;
    case SHOWDICTIONARYTOKENS:
      if (dictionary != null) {
        new ShowDictionaryTokens(dictionary);
      } else {
        printError(program);
      }
      break;
    case TRIMDICTIONARY:
      if (dictionary != null) {
        if (output == null)
          output = dictionary.substring(0, dictionary.indexOf(".")) + "_trimmed_" + min_idf + "-" + max_idf + ".xml";
        TrimDictionary trimDictionary = new TrimDictionary(min_idf, max_idf, dictionary);
        trimDictionary.trim(output);
      } else {
        printError(program);
      }
      break;
    case BITVECTORFINGERPRINTER:
      if (dictionary != null && input != null) {
        if (output == null) {
          Random random = new Random();
          output = "fingerprint_" + random.nextInt(Integer.MAX_VALUE) + ".xml";
        }
        BitVectorFingerprinter bvFingerprinter = new BitVectorFingerprinter(dictionary);
        if (config != null)
          bvFingerprinter.outputFields(config);
        String fingerprint_output = bvFingerprinter.computeFingerprintXML(input);
        bvFingerprinter.generateXML(fingerprint_output, input, output);
        System.out.println();
        System.out.println(bvFingerprinter.getFingerprintName() + " (Base64):");
        System.out.println(fingerprint_output);
      } else {
        printError(program);
      }
      break;
    case SCOREFINGERPRINTS:
      if (matcher != null && fingerprint != null && fingerprint2 != null) {
        ScoreFingerprints sfp = new ScoreFingerprints(matcher, fingerprint, fingerprint2);
        sfp.printMatch();
      } else {
        printError(program);
      }
      break;
    case EXTRACTDICTIONARY:
      if (input != null && output != null) {
        ExtractDictionary extractDictionary = new ExtractDictionary();
        extractDictionary.extractDictionaryXML(input, output);
      } else {
        printError(program);
      }
      break;
    case COMPAREDIRECTORY:
      if (matcher != null && fingerprint != null && directory != null) {
        CompareDirectory compareDirectory = new CompareDirectory(matcher, fingerprint, directory, min_score, dictionary);
        compareDirectory.computeFingerprints();
        compareDirectory.printScores();
      } else {
        printError(program);
      }
      break;
    case EXPERIMENT:
      if (config != null) {
        Experiment exp = new Experiment(config);
        exp.showSettings();
        exp.runExperiment();
      } else {
        printError(program);
      }
      break;
    default:
      break;
    }
  }

  /**
   * Provides a command line interface to run the different SDTEXT programs
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the name of the program to run<br>
   *          <code>args[...]</code> the options needed to run the program
   */
  public static void main(String args[]) {
    try {
      for (int i = 1; i < args.length; i += 2) {
        if (args[i].trim().equalsIgnoreCase("-i") || args[i].trim().equalsIgnoreCase("--input")) {
          input = args[i + 1].trim();
        } else if (args[i].trim().equalsIgnoreCase("-o") || args[i].trim().equalsIgnoreCase("--output")) {
          output = args[i + 1].trim();
        } else if (args[i].trim().equalsIgnoreCase("-c") || args[i].trim().equalsIgnoreCase("--config")) {
          config = args[i + 1].trim();
        } else if (args[i].trim().equalsIgnoreCase("-d") || args[i].trim().equalsIgnoreCase("--dictionary")) {
          dictionary = args[i + 1].trim();
        } else if (args[i].trim().equalsIgnoreCase("-p") || args[i].trim().equalsIgnoreCase("--directory")) {
          directory = args[i + 1].trim();
        } else if (args[i].trim().equalsIgnoreCase("-f") || args[i].trim().equalsIgnoreCase("--fingerprint")) {
          if (fingerprint == null) {
            fingerprint = args[i + 1].trim();
          } else if (fingerprint2 == null) {
            fingerprint2 = args[i + 1].trim();
          } else {
            ExceptionHandler.handleException(new Exception("SDTEXT: Too many fingerprints specified."));
          }
        } else if (args[i].trim().equalsIgnoreCase("-m") || args[i].trim().equalsIgnoreCase("--matcher")) {
          matcher = args[i + 1].trim();
        } else if (args[i].trim().equalsIgnoreCase("-b") || args[i].trim().equalsIgnoreCase("--min-idf")) {
          min_idf = Double.valueOf(args[i + 1]);
        } else if (args[i].trim().equalsIgnoreCase("-t") || args[i].trim().equalsIgnoreCase("--max-idf")) {
          max_idf = Double.valueOf(args[i + 1]);
        } else if (args[i].trim().equalsIgnoreCase("-s") || args[i].trim().equalsIgnoreCase("--min-score")) {
          min_score = Integer.valueOf(args[i + 1]);
        } else {
          System.err.println("SDTEXT: " + args[i] + " " + args[i + 1] + " is not a valid OPTION.");
          System.err.println("Try 'sdtext --help' for more information.");
          System.exit(-1);
        }
      }

      String program = args[0].trim();
      if (program.equalsIgnoreCase("ArabicFileTokenizer")) {
        sdtext(Program.ARABICFILETOKENIZER);
      } else if (program.equalsIgnoreCase("BitVectorFingerprinter")) {
        sdtext(Program.BITVECTORFINGERPRINTER);
      } else if (program.equalsIgnoreCase("ChineseFileTokenizer")) {
        sdtext(Program.CHINESEFILETOKENIZER);
      } else if (program.equalsIgnoreCase("OutsideInFileTokenizer")) {
        sdtext(Program.OUTSIDEINFILETOKENIZER);
      } else if (program.equalsIgnoreCase("CompareDirectory")) {
        sdtext(Program.COMPAREDIRECTORY);
      } else if (program.equalsIgnoreCase("CreateDictionary")) {
        sdtext(Program.CREATEDICTIONARY);
      } else if (program.equalsIgnoreCase("Experiment")) {
        sdtext(Program.EXPERIMENT);
      } else if (program.equalsIgnoreCase("ExtractDictionary")) {
        sdtext(Program.EXTRACTDICTIONARY);
      } else if (program.equalsIgnoreCase("ScoreFingerprints")) {
        sdtext(Program.SCOREFINGERPRINTS);
      } else if (program.equalsIgnoreCase("ShowDictionaryStatistics")) {
        sdtext(Program.SHOWDICTIONARYSTATISTICS);
      } else if (program.equalsIgnoreCase("ShowDictionaryTokens")) {
        sdtext(Program.SHOWDICTIONARYTOKENS);
      } else if (program.equalsIgnoreCase("TokenizeFile")) {
        sdtext(Program.TOKENIZEFILE);
      } else if (program.equalsIgnoreCase("TrimDictionary")) {
        sdtext(Program.TRIMDICTIONARY);
      } else if (program.equalsIgnoreCase("-h") || program.equalsIgnoreCase("--help")) {
        printHelp();
      } else {
        System.err.println("SDTEXT: " + program + " is not a valid PROGRAM.");
        System.err.println("Try 'sdtext --help' for more information.");
      }
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("SDTEXT: Incorrect number of arguments.");
      System.err.println("Try 'sdtext --help' for more information.");
    }
  }
}
